=== modified file 'Makefile.am'
--- Makefile.am	2008-08-02 00:31:54 +0000
+++ Makefile.am	2008-08-05 06:19:05 +0000
@@ -62,7 +62,6 @@
     test-ext-jp \
     test-ext-stress \
     test-ext \
-    test-embedded \
     test-fast \
     test-fast-cursor \
     test-fast-view \

=== modified file 'drizzled/handler.h'
--- drizzled/handler.h	2008-08-01 02:41:21 +0000
+++ drizzled/handler.h	2008-08-05 06:19:05 +0000
@@ -363,6 +363,7 @@
 #define HA_CREATE_USED_KEY_BLOCK_SIZE   (1L << 19)
 #define HA_CREATE_USED_TRANSACTIONAL    (1L << 20)
 #define HA_CREATE_USED_PAGE_CHECKSUM    (1L << 21)
+#define HA_CREATE_USED_BLOCK_SIZE       (1L << 22)
 
 typedef uint64_t my_xid; // this line is the same as in log_event.h
 #define MYSQL_XID_PREFIX "MySQLXid"
@@ -763,6 +764,7 @@
   uint32_t avg_row_length;
   uint32_t used_fields;
   uint32_t key_block_size;
+  uint32_t block_size;
   SQL_LIST merge_list;
   handlerton *db_type;
   enum row_type row_type;

=== modified file 'drizzled/lex.h'
--- drizzled/lex.h	2008-08-02 01:04:23 +0000
+++ drizzled/lex.h	2008-08-05 06:19:05 +0000
@@ -78,6 +78,7 @@
   { "BIT",		SYM(BIT_SYM)},
   { "BLOB",		SYM(BLOB_SYM)},
   { "BLOCK",  SYM(BLOCK_SYM)},
+  { "BLOCK_SIZE", SYM(BLOCK_SIZE_SYM)},
   { "BOOL",		SYM(BOOL_SYM)},
   { "BOOLEAN",		SYM(BOOLEAN_SYM)},
   { "BOTH",		SYM(BOTH)},

=== modified file 'drizzled/sql_show.cc'
--- drizzled/sql_show.cc	2008-08-05 02:31:51 +0000
+++ drizzled/sql_show.cc	2008-08-05 06:19:05 +0000
@@ -959,6 +959,13 @@
       end= int64_t10_to_str(table->s->key_block_size, buff, 10);
       packet->append(buff, (uint) (end - buff));
     }
+    if (share->block_size)
+    {
+      char *end;
+      packet->append(STRING_WITH_LEN(" BLOCK_SIZE="));
+      end= int64_t10_to_str(share->block_size, buff,10);
+      packet->append(buff, (uint) (end - buff));
+    }
     table->file->append_create_info(packet);
     if (share->comment.length)
     {
@@ -2814,7 +2821,7 @@
   }
   else
   {
-    char option_buff[350],*ptr;
+    char option_buff[400],*ptr;
     TABLE *show_table= tables->table;
     TABLE_SHARE *share= show_table->s;
     handler *file= show_table->file;
@@ -2868,6 +2875,12 @@
       ptr=strxmov(ptr, " row_format=", 
                   ha_row_type[(uint) share->row_type],
                   NullS);
+    if (share->block_size)
+    {
+      ptr= strmov(ptr, " block_size=");
+      ptr= int64_t10_to_str(share->block_size, ptr, 10);
+    }
+    
     if (share->transactional != HA_CHOICE_UNDEF)
     {
       ptr= strxmov(ptr, " TRANSACTIONAL=",

=== modified file 'drizzled/sql_table.cc'
--- drizzled/sql_table.cc	2008-08-05 02:31:51 +0000
+++ drizzled/sql_table.cc	2008-08-05 06:19:05 +0000
@@ -4071,6 +4071,8 @@
     create_info->max_rows= table->s->max_rows;
   if (!(used_fields & HA_CREATE_USED_AVG_ROW_LENGTH))
     create_info->avg_row_length= table->s->avg_row_length;
+  if (!(used_fields & HA_CREATE_USED_BLOCK_SIZE))
+    create_info->block_size= table->s->block_size;
   if (!(used_fields & HA_CREATE_USED_DEFAULT_CHARSET))
     create_info->default_table_charset= table->s->table_charset;
   if (!(used_fields & HA_CREATE_USED_AUTO) && table->found_next_number_field)

=== modified file 'drizzled/sql_yacc.yy'
--- drizzled/sql_yacc.yy	2008-08-05 02:31:51 +0000
+++ drizzled/sql_yacc.yy	2008-08-05 06:19:05 +0000
@@ -416,6 +416,7 @@
 %token  BIN_NUM
 %token  BIT_SYM                       /* MYSQL-FUNC */
 %token  BLOB_SYM                      /* SQL-2003-R */
+%token  BLOCK_SIZE_SYM
 %token  BLOCK_SYM
 %token  BOOLEAN_SYM                   /* SQL-2003-R */
 %token  BOOL_SYM
@@ -1483,6 +1484,11 @@
             Lex->create_info.avg_row_length=$3;
             Lex->create_info.used_fields|= HA_CREATE_USED_AVG_ROW_LENGTH;
           }
+        | BLOCK_SIZE_SYM opt_equal ulong_num    
+          { 
+            Lex->create_info.block_size= $3; 
+            Lex->create_info.used_fields|= HA_CREATE_USED_BLOCK_SIZE;
+          }
         | COMMENT_SYM opt_equal TEXT_STRING_sys
           {
             Lex->create_info.comment=$3;
@@ -6341,6 +6347,7 @@
         | AVG_SYM                  {}
         | BINLOG_SYM               {}
         | BIT_SYM                  {}
+        | BLOCK_SIZE_SYM           {}
         | BLOCK_SYM                {}
         | BOOL_SYM                 {}
         | BOOLEAN_SYM              {}

=== modified file 'drizzled/table.cc'
--- drizzled/table.cc	2008-08-05 02:31:51 +0000
+++ drizzled/table.cc	2008-08-05 06:19:05 +0000
@@ -534,6 +534,7 @@
     share->transactional= (ha_choice) (head[39] & 3);
     share->page_checksum= (ha_choice) ((head[39] >> 2) & 3);
     share->row_type= (row_type) head[40];
+    share->block_size= uint4korr(head+43);
     share->table_charset= get_charset((uint) head[38],MYF(0));
     share->null_field_first= 1;
   }
@@ -2094,7 +2095,8 @@
     /* Next few bytes where for RAID support */
     fileinfo[41]= 0;
     fileinfo[42]= 0;
-    fileinfo[43]= 0;
+    int4store(fileinfo+43,create_info->block_size);
+ 
     fileinfo[44]= 0;
     fileinfo[45]= 0;
     fileinfo[46]= 0;
@@ -2137,6 +2139,7 @@
   create_info->min_rows= share->min_rows;
   create_info->table_options= share->db_create_options;
   create_info->avg_row_length= share->avg_row_length;
+  create_info->block_size= share->block_size;
   create_info->row_type= share->row_type;
   create_info->default_table_charset= share->table_charset;
   create_info->table_charset= 0;

=== modified file 'drizzled/table.h'
--- drizzled/table.h	2008-08-01 17:09:24 +0000
+++ drizzled/table.h	2008-08-05 06:19:05 +0000
@@ -229,6 +229,7 @@
   key_map keys_for_keyread;
   ha_rows min_rows, max_rows;		/* create information */
   ulong   avg_row_length;		/* create information */
+  ulong   block_size;                   /* create information */
   ulong   version, mysql_version;
   ulong   timestamp_offset;		/* Set to offset+1 of record */
   ulong   reclength;			/* Recordlength */

=== modified file 'storage/heap/Makefile.am'
--- storage/heap/Makefile.am	2008-07-30 19:17:49 +0000
+++ storage/heap/Makefile.am	2008-08-05 05:01:04 +0000
@@ -29,7 +29,9 @@
 			hp_rrnd.c hp_scan.c hp_update.c hp_write.c hp_delete.c \
 			hp_rsame.c hp_create.c hp_rename.c hp_rfirst.c \
 			hp_rnext.c hp_rlast.c hp_rprev.c hp_clear.c \
+			hp_dspace.c hp_record.c \
 			hp_rkey.c hp_block.c \
+			hp_rectest.c \
 			ha_heap.cc \
 			hp_hash.c hp_static.c
 

=== modified file 'storage/heap/ha_heap.cc'
--- storage/heap/ha_heap.cc	2008-08-05 02:31:51 +0000
+++ storage/heap/ha_heap.cc	2008-08-05 06:19:05 +0000
@@ -391,6 +391,14 @@
 }
 
 
+enum row_type ha_heap::get_row_type() const
+{
+  if (file->s->recordspace.is_variable_size)
+    return ROW_TYPE_DYNAMIC;
+
+  return ROW_TYPE_FIXED;
+}
+
 int ha_heap::extra(enum ha_extra_function operation)
 {
   return heap_extra(file,operation);
@@ -592,22 +600,61 @@
 int ha_heap::create(const char *name, TABLE *table_arg,
 		    HA_CREATE_INFO *create_info)
 {
-  uint key, parts, mem_per_row= 0, keys= table_arg->s->keys;
+  uint key, parts, mem_per_row_keys= 0, keys= table_arg->s->keys;
   uint auto_key= 0, auto_key_type= 0;
-  ha_rows max_rows;
+  uint max_key_fieldnr = 0, key_part_size = 0, next_field_pos = 0;
+  uint column_idx, column_count= table_arg->s->fields;
+  HP_COLUMNDEF *columndef;
   HP_KEYDEF *keydef;
   HA_KEYSEG *seg;
+  char buff[FN_REFLEN];
   int error;
   TABLE_SHARE *share= table_arg->s;
   bool found_real_auto_increment= 0;
 
+  if (!(columndef= (HP_COLUMNDEF*) my_malloc(column_count * sizeof(HP_COLUMNDEF), MYF(MY_WME))))
+    return my_errno;
+
+  for (column_idx= 0; column_idx < column_count; column_idx++)
+  {
+    Field* field= *(table_arg->field + column_idx);
+    HP_COLUMNDEF* column= columndef + column_idx;
+    column->type= (uint16_t)field->type();
+    column->length= field->pack_length();
+    column->offset= field->offset(field->table->record[0]);
+
+    if (field->null_bit)
+    {
+      column->null_bit= field->null_bit;
+      column->null_pos= (uint) (field->null_ptr - (uchar*) table_arg->record[0]);
+    }
+    else
+    {
+      column->null_bit= 0;
+      column->null_pos= 0;
+    }
+
+    if (field->type() == DRIZZLE_TYPE_VARCHAR)
+    {
+      column->length_bytes= (uint8_t)(((Field_varstring*)field)->length_bytes);
+    }
+    else
+    {
+      column->length_bytes= 0;
+    }
+  }
+
   for (key= parts= 0; key < keys; key++)
     parts+= table_arg->key_info[key].key_parts;
 
   if (!(keydef= (HP_KEYDEF*) my_malloc(keys * sizeof(HP_KEYDEF) +
 				       parts * sizeof(HA_KEYSEG),
 				       MYF(MY_WME))))
+  {
+    my_free((void *) columndef, MYF(0));
     return my_errno;
+  }
+
   seg= my_reinterpret_cast(HA_KEYSEG*) (keydef + keys);
   for (key= 0; key < keys; key++)
   {
@@ -623,11 +670,11 @@
     case HA_KEY_ALG_UNDEF:
     case HA_KEY_ALG_HASH:
       keydef[key].algorithm= HA_KEY_ALG_HASH;
-      mem_per_row+= sizeof(char*) * 2; // = sizeof(HASH_INFO)
+      mem_per_row_keys+= sizeof(char*) * 2; // = sizeof(HASH_INFO)
       break;
     case HA_KEY_ALG_BTREE:
       keydef[key].algorithm= HA_KEY_ALG_BTREE;
-      mem_per_row+=sizeof(TREE_ELEMENT)+pos->key_length+sizeof(char*);
+      mem_per_row_keys+=sizeof(TREE_ELEMENT)+pos->key_length+sizeof(char*);
       break;
     default:
       assert(0); // cannot happen
@@ -652,6 +699,16 @@
       seg->length=  (uint) key_part->length;
       seg->flag=    key_part->key_part_flag;
 
+      next_field_pos= seg->start + seg->length;
+      if (field->type() == DRIZZLE_TYPE_VARCHAR)
+      {
+        next_field_pos+= (uint8_t)(((Field_varstring*)field)->length_bytes);
+      }
+
+      if (next_field_pos > key_part_size) {
+        key_part_size= next_field_pos;
+      }
+
       if (field->flags & (ENUM_FLAG | SET_FLAG))
         seg->charset= &my_charset_bin;
       else
@@ -677,11 +734,22 @@
         auto_key= key+ 1;
 	auto_key_type= field->key_type();
       }
+      if ((uint)field->field_index + 1 > max_key_fieldnr)
+      {
+        /* Do not use seg->fieldnr as it's not reliable in case of temp tables */
+        max_key_fieldnr= field->field_index + 1;
+      }
     }
   }
-  mem_per_row+= MY_ALIGN(share->reclength + 1, sizeof(char*));
-  max_rows = (ha_rows) (table_arg->in_use->variables.max_heap_table_size /
-			(uint64_t) mem_per_row);
+  
+  if (key_part_size < share->null_bytes + ((share->last_null_bit_pos+7) >> 3))
+  {
+    /* Make sure to include null fields regardless of the presense of keys */
+    key_part_size = share->null_bytes + ((share->last_null_bit_pos+7) >> 3);
+  }
+
+  
+  
   if (table_arg->found_next_number_field)
   {
     keydef[share->next_number_index].flag|= HA_AUTO_KEY;
@@ -695,14 +763,19 @@
   hp_create_info.max_table_size=current_thd->variables.max_heap_table_size;
   hp_create_info.with_auto_increment= found_real_auto_increment;
   hp_create_info.internal_table= internal_table;
-  max_rows = (ha_rows) (hp_create_info.max_table_size / mem_per_row);
-  error= heap_create(name,
-		     keys, keydef, share->reclength,
-		     (ulong) ((share->max_rows < max_rows &&
-			       share->max_rows) ? 
-			      share->max_rows : max_rows),
-		     (ulong) share->min_rows, &hp_create_info, &internal_share);
+  hp_create_info.max_chunk_size= share->block_size;
+  hp_create_info.is_dynamic= (share->row_type == ROW_TYPE_DYNAMIC);
+  error= heap_create(fn_format(buff,name,"","",
+                               MY_REPLACE_EXT|MY_UNPACK_FILENAME),
+                   keys, keydef,
+         column_count, columndef,
+         max_key_fieldnr, key_part_size,
+         share->reclength, mem_per_row_keys,
+         (ulong) share->max_rows, (ulong) share->min_rows,
+         &hp_create_info, &internal_share);
+  
   my_free((uchar*) keydef, MYF(0));
+  my_free((void *) columndef, MYF(0));
   assert(file == 0);
   return (error);
 }
@@ -713,6 +786,13 @@
   table->file->info(HA_STATUS_AUTO);
   if (!(create_info->used_fields & HA_CREATE_USED_AUTO))
     create_info->auto_increment_value= stats.auto_increment_value;
+  if (!(create_info->used_fields & HA_CREATE_USED_BLOCK_SIZE))
+  {
+    if (file->s->recordspace.is_variable_size)
+      create_info->block_size= file->s->recordspace.chunk_length;
+    else
+      create_info->block_size= 0;
+  }
 }
 
 void ha_heap::get_auto_increment(uint64_t offset __attribute__((unused)),

=== modified file 'storage/heap/ha_heap.h'
--- storage/heap/ha_heap.h	2008-07-26 18:02:40 +0000
+++ storage/heap/ha_heap.h	2008-08-05 05:01:04 +0000
@@ -45,8 +45,7 @@
     return ((table_share->key_info[inx].algorithm == HA_KEY_ALG_BTREE) ?
             "BTREE" : "HASH");
   }
-  /* Rows also use a fixed-size format */
-  enum row_type get_row_type() const { return ROW_TYPE_FIXED; }
+  enum row_type get_row_type() const;
   const char **bas_ext() const;
   uint64_t table_flags() const
   {

=== modified file 'storage/heap/heap.h'
--- storage/heap/heap.h	2008-07-29 02:48:34 +0000
+++ storage/heap/heap.h	2008-08-05 05:01:04 +0000
@@ -127,22 +127,51 @@
   uint (*get_key_length)(struct st_hp_keydef *keydef, const uchar *key);
 } HP_KEYDEF;
 
+typedef struct st_heap_columndef              /* column information */
+{
+  int16_t  type;                                /* en_fieldtype */
+  uint32_t length;                      /* length of field */
+  uint32_t offset;                      /* Offset to position in row */
+  uint8_t  null_bit;                    /* If column may be 0 */
+  uint16_t null_pos;                    /* position for null marker */
+  uint8_t  length_bytes;  /* length of the size, 1 o 2 bytes */
+} HP_COLUMNDEF;
+ 
+typedef struct st_heap_dataspace   /* control data for data space */
+{
+  HP_BLOCK block;
+  uint chunk_count;             /* Total chunks ever allocated in this dataspace */
+  uint del_chunk_count;         /* Deleted chunks count */
+  uchar *del_link;               /* Link to last deleted chunk */
+  uint chunk_length;            /* Total length of one chunk */
+  uint chunk_dataspace_length;  /* Length of payload that will be placed into one chunk */
+  uint offset_status;           /* Offset of the status flag relative to the chunk start */
+  uint offset_link;             /* Offset of the linking pointer relative to the chunk start */
+  uint is_variable_size;          /* Test whether records have variable size and so "next" pointer */
+  uint64_t total_data_length;  /* Total size allocated within this data space */
+} HP_DATASPACE;
+
+
 typedef struct st_heap_share
 {
-  HP_BLOCK block;
   HP_KEYDEF  *keydef;
+  HP_COLUMNDEF *column_defs;
+  HP_DATASPACE recordspace;  /* Describes "block", which contains actual records */
+
   ulong min_records,max_records;	/* Params to open */
-  uint64_t data_length,index_length,max_table_size;
+  uint64_t index_length,max_table_size;
   uint key_stat_version;                /* version to indicate insert/delete */
-  uint records;				/* records */
-  uint blength;				/* records rounded up to 2^n */
-  uint deleted;				/* Deleted records in database */
-  uint reclength;			/* Length of one record */
+  uint records;             /* Actual record (row) count */
+  uint blength;                                     /* used_chunk_count rounded up to 2^n */
+  uint fixed_data_length;     /* Length of record's fixed part, which contains keys and always fits into the first chunk */
+  uint fixed_column_count;  /* Number of columns stored in fixed_data_length */
   uint changed;
   uint keys,max_key_length;
+  uint column_count;
   uint currently_disabled_keys;    /* saved value from "keys" when disabled */
   uint open_count;
-  uchar *del_link;			/* Link to next block with del. rec */
+
+  
   char * name;			/* Name of "memory-file" */
   THR_LOCK lock;
   pthread_mutex_t intern_lock;		/* Locking for use with _locking */
@@ -180,6 +209,8 @@
 {
   uint auto_key;                        /* keynr [1 - maxkey] for auto key */
   uint auto_key_type;
+  uint max_chunk_size;
+  uint is_dynamic;  
   uint64_t max_table_size;
   uint64_t auto_increment;
   my_bool with_auto_increment;
@@ -192,16 +223,20 @@
 extern HP_INFO *heap_open_from_share(HP_SHARE *share, int mode);
 extern HP_INFO *heap_open_from_share_and_register(HP_SHARE *share, int mode);
 extern int heap_close(HP_INFO *info);
-extern int heap_write(HP_INFO *info,const uchar *buff);
-extern int heap_update(HP_INFO *info,const uchar *old,const uchar *newdata);
+extern int heap_write(HP_INFO *info,const uchar *record);
+extern int heap_update(HP_INFO *info,const uchar *old_record,const uchar *new_record);
 extern int heap_rrnd(HP_INFO *info,uchar *buf,uchar *pos);
 extern int heap_scan_init(HP_INFO *info);
 extern int heap_scan(register HP_INFO *info, uchar *record);
 extern int heap_delete(HP_INFO *info,const uchar *buff);
 extern int heap_info(HP_INFO *info,HEAPINFO *x,int flag);
 extern int heap_create(const char *name, uint keys, HP_KEYDEF *keydef,
-		       uint reclength, ulong max_records, ulong min_records,
-		       HP_CREATE_INFO *create_info, HP_SHARE **share);
+           uint columns, HP_COLUMNDEF *columndef,
+           uint max_key_fieldnr, uint key_part_size,
+           uint reclength, uint keys_memory_size,
+           ulong max_records, ulong min_records,
+           HP_CREATE_INFO *create_info, HP_SHARE **share);
+
 extern int heap_delete_table(const char *name);
 extern void heap_drop_table(HP_INFO *info);
 extern int heap_extra(HP_INFO *info,enum ha_extra_function function);

=== modified file 'storage/heap/heapdef.h'
--- storage/heap/heapdef.h	2008-07-30 05:33:20 +0000
+++ storage/heap/heapdef.h	2008-08-05 05:01:04 +0000
@@ -32,6 +32,10 @@
 #define HP_MIN_RECORDS_IN_BLOCK 16
 #define HP_MAX_RECORDS_IN_BLOCK 8192
 
+#define CHUNK_STATUS_DELETED 0    /* this chunk has been deleted and can be reused */
+#define CHUNK_STATUS_ACTIVE  1    /* this chunk represents the first part of a live record */
+#define CHUNK_STATUS_LINKED  2    /* this chunk is a continuation from another chunk (part of chunkset) */
+
 	/* Some extern variables */
 
 extern LIST *heap_open_list,*heap_share_list;
@@ -42,7 +46,11 @@
 #define hp_find_hash(A,B) ((HASH_INFO*) hp_find_block((A),(B)))
 
 	/* Find pos for record and update it in info->current_ptr */
-#define hp_find_record(info,pos) (info)->current_ptr= hp_find_block(&(info)->s->block,pos)
+#define hp_find_record(info,pos) (info)->current_ptr= hp_find_block(&(info)->s->recordspace.block,pos)
+
+#define get_chunk_status(info,ptr) (ptr[(info)->offset_status])
+
+#define get_chunk_count(info,rec_length) ((rec_length + (info)->chunk_dataspace_length - 1) / (info)->chunk_dataspace_length)
 
 typedef struct st_hp_hash_info
 {
@@ -100,5 +108,20 @@
 extern void hp_clear_keys(HP_SHARE *info);
 extern uint hp_rb_pack_key(HP_KEYDEF *keydef, uchar *key, const uchar *old,
                            key_part_map keypart_map);
+
+   /* Chunkset management (alloc/free/encode/decode) functions */
+ 
+extern uchar *hp_allocate_chunkset(HP_DATASPACE *info, uint chunk_count);
+extern int hp_reallocate_chunkset(HP_DATASPACE *info, uint chunk_count, uchar* pos);
+extern void hp_free_chunks(HP_DATASPACE *info, uchar *pos);
+extern void hp_clear_dataspace(HP_DATASPACE *info);
+ 
+extern uint hp_get_encoded_data_length(HP_SHARE *info, const uchar *record, uint *chunk_count);
+extern void hp_copy_record_data_to_chunkset(HP_SHARE *info, const uchar *record, uchar *pos);
+extern void hp_extract_record(HP_SHARE *info, uchar *record, const uchar *pos);
+extern uint hp_process_record_data_to_chunkset(HP_SHARE *info, const uchar *record, uchar *pos, uint is_compare);
+
+
+
 extern pthread_mutex_t THR_LOCK_heap;
 C_MODE_END

=== modified file 'storage/heap/hp_clear.c'
--- storage/heap/hp_clear.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_clear.c	2008-08-05 05:01:04 +0000
@@ -28,16 +28,11 @@
 
 void hp_clear(HP_SHARE *info)
 {
-  if (info->block.levels)
-    VOID(hp_free_level(&info->block,info->block.levels,info->block.root,
-			(uchar*) 0));
-  info->block.levels=0;
+  hp_clear_dataspace(&info->recordspace);
   hp_clear_keys(info);
-  info->records= info->deleted= 0;
-  info->data_length= 0;
+  info->records= 0;
   info->blength=1;
   info->changed=0;
-  info->del_link=0;
   return;
 }
 
@@ -154,7 +149,7 @@
   int error= 0;
   HP_SHARE *share= info->s;
 
-  if (share->data_length || share->index_length)
+  if (share->recordspace.total_data_length || share->index_length)
     error= HA_ERR_CRASHED;
   else
     if (share->currently_disabled_keys)
@@ -188,4 +183,3 @@
 
   return (! share->keys && share->currently_disabled_keys);
 }
-

=== modified file 'storage/heap/hp_create.c'
--- storage/heap/hp_create.c	2008-08-02 19:53:06 +0000
+++ storage/heap/hp_create.c	2008-08-05 06:19:05 +0000
@@ -14,18 +14,32 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
 
 #include "heapdef.h"
+#include "libdrizzle/drizzle_com.h"
+#include "drizzled/error.h"
 
 static int keys_compare(heap_rb_param *param, uchar *key1, uchar *key2);
-static void init_block(HP_BLOCK *block,uint reclength,ulong min_records,
-		       ulong max_records);
+static void init_block(HP_BLOCK *block,uint chunk_length, ulong min_records,
+                        ulong max_records);
+
+#define FIXED_REC_OVERHEAD (sizeof(uchar))
+#define VARIABLE_REC_OVERHEAD (sizeof(uchar**) + sizeof(uchar))
+ 
+/* Minimum size that a chunk can take, 12 bytes on 32bit, 24 bytes on 64bit */
+#define VARIABLE_MIN_CHUNK_SIZE \
+        ((sizeof(uchar**) + VARIABLE_REC_OVERHEAD + sizeof(uchar**) - 1) & ~(sizeof(uchar**) - 1))
+
 
 /* Create a heap table */
 
 int heap_create(const char *name, uint keys, HP_KEYDEF *keydef,
-		uint reclength, ulong max_records, ulong min_records,
-		HP_CREATE_INFO *create_info, HP_SHARE **res)
+    uint columns, HP_COLUMNDEF *columndef,
+    uint max_key_fieldnr, uint key_part_size,
+    uint reclength, uint keys_memory_size,
+    ulong max_records, ulong min_records,
+    HP_CREATE_INFO *create_info, HP_SHARE **res)
 {
   uint i, j, key_segs, max_length, length;
+  ulong max_rows_for_stated_memory;
   HP_SHARE *share= 0;
   HA_KEYSEG *keyseg;
 
@@ -41,13 +55,124 @@
 
   if (!share)
   {
+    uint chunk_dataspace_length, chunk_length, is_variable_size;
+    uint fixed_data_length, fixed_column_count;
     HP_KEYDEF *keyinfo;
-    
+
+    if (create_info->max_chunk_size)
+    {
+      uint configured_chunk_size= create_info->max_chunk_size;
+
+      /* User requested variable-size records, let's see if they're possible */
+
+      if (configured_chunk_size < key_part_size)
+      {
+        /* Eventual chunk_size cannot be smaller than key data,
+           which allows all keys to fit into the first chunk */
+        my_error(ER_CANT_USE_OPTION_HERE, MYF(0), "block_size");
+        pthread_mutex_unlock(&THR_LOCK_heap);
+        return(ER_CANT_USE_OPTION_HERE);
+      }
+
+      if ((reclength - configured_chunk_size) >= VARIABLE_MIN_CHUNK_SIZE<<1)
+      {
+        /* Allow variable size only if we're saving at least two smallest chunks */
+        /* There has to be at least one field after indexed fields */
+        /* Note that NULL bits are already included in key_part_size */
+        is_variable_size= 1;
+        chunk_dataspace_length= configured_chunk_size;
+      }
+      else
+      {
+        /* max_chunk_size is near the full reclength, let's use fixed size */
+        is_variable_size= 0;
+        chunk_dataspace_length= reclength;
+      }
+    }
+    else if (create_info->is_dynamic)
+    {
+      /* User asked for dynamic records - use 256 as the chunk size */
+      if ((key_part_size + VARIABLE_REC_OVERHEAD) > 256)
+        chunk_dataspace_length= key_part_size;
+      else
+        chunk_dataspace_length= 256 - VARIABLE_REC_OVERHEAD;
+
+      is_variable_size= 1;
+    }
+    else
+    {
+      /* if max_chunk_size is not specified, put the whole record in one chunk */
+      is_variable_size= 0;
+      chunk_dataspace_length= reclength;
+    }
+
+    if (is_variable_size)
+    {
+      /* Check whether we have any variable size records past key data */
+      uint has_variable_fields= 0;
+
+      fixed_data_length= key_part_size;
+      fixed_column_count= max_key_fieldnr;
+
+      for (i= max_key_fieldnr; i < columns; i++)
+      {
+        HP_COLUMNDEF* column= columndef + i;
+        if (column->type == DRIZZLE_TYPE_VARCHAR && column->length >= 32)
+        {
+            /* The field has to be >= 5.0.3 true VARCHAR and have substantial length */
+            /* TODO: do we want to calculate minimum length? */
+            has_variable_fields= 1;
+            break;
+        }
+
+        if (has_variable_fields)
+        {
+          break;
+        }
+
+        if ((column->offset + column->length) <= chunk_dataspace_length)
+        {
+          /* Still no variable-size columns, add one fixed-length */
+          fixed_column_count= i + 1;
+          fixed_data_length= column->offset + column->length;
+        }
+      }
+
+      if (!has_variable_fields)
+      {
+        /* There is no need to use variable-size records without variable-size columns */
+        /* Reset sizes if it's not variable size anymore */
+        is_variable_size= 0;
+        chunk_dataspace_length= reclength;
+        fixed_data_length= reclength;
+        fixed_column_count= columns;
+      }
+    }
+    else
+    {
+      fixed_data_length= reclength;
+      fixed_column_count= columns;
+    }
+
     /*
-      We have to store sometimes uchar* del_link in records,
-      so the record length should be at least sizeof(uchar*)
+      We store uchar* del_link inside the data area of deleted records,
+      so the data length should be at least sizeof(uchar*)
     */
-    set_if_bigger(reclength, sizeof (uchar*));
+    set_if_bigger(chunk_dataspace_length, sizeof (uchar**));
+
+    if (is_variable_size)
+    {
+      chunk_length= chunk_dataspace_length + VARIABLE_REC_OVERHEAD;
+    }
+    else
+    {
+      chunk_length= chunk_dataspace_length + FIXED_REC_OVERHEAD;
+    }
+
+    /* Align chunk length to the next pointer */
+    chunk_length= (uint) (chunk_length + sizeof(uchar**) - 1) & ~(sizeof(uchar**) - 1);
+
+
     
     for (i= key_segs= max_length= 0, keyinfo= keydef; i < keys; i++, keyinfo++)
     {
@@ -125,14 +250,29 @@
     }
     if (!(share= (HP_SHARE*) my_malloc((uint) sizeof(HP_SHARE)+
 				       keys*sizeof(HP_KEYDEF)+
+                                       columns*sizeof(HP_COLUMNDEF)+
 				       key_segs*sizeof(HA_KEYSEG),
 				       MYF(MY_ZEROFILL))))
       goto err;
-    share->keydef= (HP_KEYDEF*) (share + 1);
+
+    /*
+       Max_records is used for estimating block sizes and for enforcement.
+       Calculate the very maximum number of rows (if everything was one chunk) and
+       then take either that value or configured max_records (pick smallest one)
+    */
+    max_rows_for_stated_memory= (ha_rows) (create_info->max_table_size /
+      (keys_memory_size + chunk_length));
+    max_records = ((max_records && max_records < max_rows_for_stated_memory) ? 
+                      max_records : max_rows_for_stated_memory);
+ 
+    share->column_defs= (HP_COLUMNDEF*) (share + 1);
+    memcpy(share->column_defs, columndef, (size_t) (sizeof(columndef[0]) * columns));
+
+    share->keydef= (HP_KEYDEF*) (share->column_defs + columns);    
     share->key_stat_version= 1;
     keyseg= (HA_KEYSEG*) (share->keydef + keys);
-    init_block(&share->block, reclength + 1, min_records, max_records);
-	/* Fix keys */
+    init_block(&share->recordspace.block, chunk_length, min_records, max_records);
+    /* Fix keys */
     memcpy(share->keydef, keydef, (size_t) (sizeof(keydef[0]) * keys));
     for (i= 0, keyinfo= share->keydef; i < keys; i++, keyinfo++)
     {
@@ -169,15 +309,32 @@
     share->min_records= min_records;
     share->max_records= max_records;
     share->max_table_size= create_info->max_table_size;
-    share->data_length= share->index_length= 0;
-    share->reclength= reclength;
+    share->index_length= 0;
     share->blength= 1;
     share->keys= keys;
     share->max_key_length= max_length;
+    share->column_count= columns;
     share->changed= 0;
     share->auto_key= create_info->auto_key;
     share->auto_key_type= create_info->auto_key_type;
     share->auto_increment= create_info->auto_increment;
+
+    share->fixed_data_length= fixed_data_length;
+    share->fixed_column_count= fixed_column_count;
+
+    share->recordspace.chunk_length= chunk_length;
+    share->recordspace.chunk_dataspace_length= chunk_dataspace_length;
+    share->recordspace.is_variable_size= is_variable_size;
+    share->recordspace.total_data_length= 0;
+
+    if (is_variable_size) {
+      share->recordspace.offset_link= chunk_dataspace_length;
+      share->recordspace.offset_status= share->recordspace.offset_link + sizeof(uchar**);
+    } else {
+      share->recordspace.offset_link= 1<<22; /* Make it likely to fail if anyone uses this offset */
+      share->recordspace.offset_status= chunk_dataspace_length;
+    }
+
     /* Must be allocated separately for rename to work */
     if (!(share->name= my_strdup(name,MYF(0))))
     {
@@ -214,7 +371,7 @@
 		    param->search_flag, not_used);
 }
 
-static void init_block(HP_BLOCK *block, uint reclength, ulong min_records,
+static void init_block(HP_BLOCK *block, uint chunk_length, ulong min_records,
 		       ulong max_records)
 {
   uint i,recbuffer,records_in_block;
@@ -222,7 +379,9 @@
   max_records= max(min_records,max_records);
   if (!max_records)
     max_records= 1000;			/* As good as quess as anything */
-  recbuffer= (uint) (reclength + sizeof(uchar**) - 1) & ~(sizeof(uchar**) - 1);
+  
+  /* we want to start each chunk at 8 bytes boundary, round recbuffer to the next 8 */
+  recbuffer= (uint) (chunk_length + sizeof(uchar**) - 1) & ~(sizeof(uchar**) - 1);  
   records_in_block= max_records / 10;
   if (records_in_block < 10 && max_records)
     records_in_block= 10;

=== modified file 'storage/heap/hp_delete.c'
--- storage/heap/hp_delete.c	2008-07-17 16:34:04 +0000
+++ storage/heap/hp_delete.c	2008-08-05 05:01:04 +0000
@@ -22,6 +22,7 @@
   uchar *pos;
   HP_SHARE *share=info->s;
   HP_KEYDEF *keydef, *end, *p_lastinx;
+  uint rec_length, chunk_count;
 
   test_active(info);
 
@@ -29,6 +30,8 @@
     return(my_errno);			/* Record changed */
   share->changed=1;
 
+  rec_length = hp_get_encoded_data_length(share, record, &chunk_count);
+
   if ( --(share->records) < share->blength >> 1) share->blength>>=1;
   pos=info->current_ptr;
 
@@ -41,10 +44,7 @@
   }
 
   info->update=HA_STATE_DELETED;
-  *((uchar**) pos)=share->del_link;
-  share->del_link=pos;
-  pos[share->reclength]=0;		/* Record deleted */
-  share->deleted++;
+  hp_free_chunks(&share->recordspace, pos);
   info->current_hash_ptr=0;
 
   return(0);
@@ -106,6 +106,9 @@
   blength=share->blength;
   if (share->records+1 == blength)
     blength+= blength;
+
+  /* find the very last HASH_INFO pointer in the index */
+  /* note that records has already been decremented */
   lastpos=hp_find_hash(&keyinfo->block,share->records);
   last_ptr=0;
 
@@ -135,16 +138,24 @@
     info->current_ptr = last_ptr ? last_ptr->ptr_to_rec : 0;
   }
   empty=pos;
-  if (gpos)
-    gpos->next_key=pos->next_key;	/* unlink current ptr */
+  if (gpos) {
+    /* gpos says we have previous HASH_INFO, change previous to point to next, this way unlinking "empty" */
+    gpos->next_key=pos->next_key;
+  }
   else if (pos->next_key)
   {
+    /* no previous gpos, this pos is the first in the list and it has pointer to "next" */
+    /* move next HASH_INFO data to our pos, to free up space at the next position */
+    /* remember next pos as "empty", nobody refers to "empty" at this point */
     empty=pos->next_key;
     pos->ptr_to_rec=empty->ptr_to_rec;
     pos->next_key=empty->next_key;
   }
   else
+  {
+    /* this was the only HASH_INFO at this position */
     keyinfo->hash_buckets--;
+  }
 
   if (empty == lastpos)			/* deleted last hash key */
     return (0);

=== added file 'storage/heap/hp_dspace.c'
--- storage/heap/hp_dspace.c	1970-01-01 00:00:00 +0000
+++ storage/heap/hp_dspace.c	2008-08-05 05:01:04 +0000
@@ -0,0 +1,419 @@
+/* Copyright (C) 2000-2002 MySQL AB
+   Copyright (C) 2008 eBay, Inc
+
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+
+/* Implements various base dataspace-related functions - allocate, free, clear */
+
+#include "heapdef.h"
+
+
+/*
+  MySQL Heap tables keep data in arrays of fixed-size chunks.
+  These chunks are organized into two groups of HP_BLOCK structures:
+    - group1 contains indexes, with one HP_BLOCK per key
+      (part of HP_KEYDEF)
+    - group2 contains record data, with single HP_BLOCK
+      for all records, referenced by HP_SHARE.recordspace.block
+
+  While columns used in index are usually small, other columns
+  in the table may need to accomodate larger data. Typically,
+  larger data is placed into VARCHAR or BLOB columns. With actual
+  sizes varying, Heap Engine has to support variable-sized records
+  in memory. Heap Engine implements the concept of dataspace
+  (HP_DATASPACE), which incorporates HP_BLOCK for the record data,
+  and adds more information for managing variable-sized records.
+
+  Variable-size records are stored in multiple "chunks",
+  which means that a single record of data (database "row") can
+  consist of multiple chunks organized into one "set". HP_BLOCK
+  contains chunks. In variable-size format, one record
+  is represented as one or many chunks, depending on the actual
+  data, while in fixed-size mode, one record is always represented
+  as one chunk. The index structures would always point to the first
+  chunk in the chunkset.
+
+  At the time of table creation, Heap Engine attempts to find out
+  if variable-size records are desired. A user can request
+  variable-size records by providing either row_type=dynamic or
+  block_size=NNN table create option. Heap Engine will check
+  whether block_size provides enough space in the first chunk
+  to keep all null bits and columns that are used in indexes.
+  If block_size is too small, table creation will be aborted
+  with an error. Heap Engine will revert to fixed-size allocation
+  mode if block_size provides no memory benefits (no VARCHAR
+  fields extending past first chunk).
+
+  In order to improve index search performance, Heap Engine needs
+  to keep all null flags and all columns used as keys inside
+  the first chunk of a chunkset. In particular, this means that
+  all columns used as keys should be defined first in the table
+  creation SQL. The length of data used by null bits and key columns
+  is stored as fixed_data_length inside HP_SHARE. fixed_data_length
+  will extend past last key column if more fixed-length fields can
+  fit into the first chunk.
+
+  Variable-size records are necessary only in the presence
+  of variable-size columns. Heap Engine will be looking for VARCHAR
+  columns, which declare length of 32 or more. If no such columns
+  are found, table will be switched to fixed-size format. You should
+  always try to put such columns at the end of the table definition.
+
+  Whenever data is being inserted or updated in the table
+  Heap Engine will calculate how many chunks are necessary.
+  For insert operations, Heap Engine allocates new chunkset in
+  the recordspace. For update operations it will modify length of
+  the existing chunkset, unlinking unnecessary chunks at the end,
+  or allocating and adding more if larger length is necessary.
+
+  When writing data to chunks or copying data back to record,
+  Heap Engine will first copy fixed_data_length of data using single
+  memcpy call. The rest of the columns are processed one-by-one.
+  Non-VARCHAR columns are copied in their full format. VARCHAR's
+  are copied based on their actual length. Any NULL values after
+  fixed_data_length are skipped.
+
+  The allocation and contents of the actual chunks varies between
+  fixed and variable-size modes. Total chunk length is always
+  aligned to the next sizeof(uchar*). Here is the format of
+  fixed-size chunk:
+      uchar[] - sizeof=chunk_dataspace_length, but at least
+               sizeof(uchar*) bytes. Keeps actual data or pointer
+               to the next deleted chunk.
+               chunk_dataspace_length equals to full record length
+      uchar   - status field (1 means "in use", 0 means "deleted")
+  Variable-size uses different format:
+      uchar[] - sizeof=chunk_dataspace_length, but at least
+               sizeof(uchar*) bytes. Keeps actual data or pointer
+               to the next deleted chunk.
+               chunk_dataspace_length is set according to table
+               setup (block_size)
+      uchar*  - pointer to the next chunk in this chunkset,
+               or NULL for the last chunk
+      uchar  -  status field (1 means "first", 0 means "deleted",
+               2 means "linked")
+
+  When allocating a new chunkset of N chunks, Heap Engine will try
+  to allocate chunks one-by-one, linking them as they become
+  allocated. Allocation of a single chunk will attempt to reuse
+  a deleted (freed) chunk. If no free chunks are available,
+  it will attempt to allocate a new area inside HP_BLOCK.
+  Freeing chunks will place them at the front of free list
+  referenced by del_link in HP_DATASPACE. The newly freed chunk
+  will contain reference to the previously freed chunk in its first
+  sizeof(uchar*) of the payload space.
+
+  Here is open issues:
+    1. It is not very nice to require people to keep key columns
+       at the beginning of the table creation SQL. There are three
+       proposed resolutions:
+       a. Leave it as is. It's a reasonable limitation
+       b. Add new HA_KEEP_KEY_COLUMNS_TO_FRONT flag to handler.h and
+          make table.cpp align columns when it creates the table
+       c. Make HeapEngine reorder columns in the chunk data, so that
+          key columns go first. Add parallel HA_KEYSEG structures
+          to distinguish positions in record vs. positions in
+          the first chunk. Copy all data field-by-field rather than
+          using single memcpy unless DBA kept key columns to
+          the beginning.
+    2. heap_check_heap needs verify linked chunks, looking for
+       issues such as orphans, cycles, and bad links. However,
+       Heap Engine today does not do similar things even for
+       free list.
+    3. With new HP_DATASPACE allocation mechaism, BLOB will become
+       increasingly simple to implement, but I may not have time
+       for that. In one approach, BLOB data can be placed at
+       the end of the same record. In another approach (which I
+       prefer) BLOB data would have its own HP_DATASPACE with
+       variable-size entries.
+    4. In a more sophisticated implementation, some space can
+       be saved even with all fixed-size columns if many of them
+       have NULL value, as long as these columns are not used
+       in indexes
+    5. In variable-size format status should be moved to lower
+       bits of the "next" pointer. Pointer is always aligned
+       to sizeof(uchar*), which is at least 4, leaving 2 lower
+       bits free. This will save 8 bytes per chunk
+       on 64-bit platform.
+    6. As we do not want to modify FRM format, BLOCK_SIZE option
+       of "CREATE TABLE" is saved as "RAID_CHUNKSIZE" for
+       Heap Engine tables.
+*/
+
+static uchar *hp_allocate_one_chunk(HP_DATASPACE *info);
+
+
+/**
+  Clear a dataspace
+
+  Frees memory and zeros-out any relevant counters in the dataspace
+
+  @param  info  the dataspace to clear
+*/
+
+void hp_clear_dataspace(HP_DATASPACE *info)
+{
+  if (info->block.levels)
+  {
+    VOID(hp_free_level(&info->block,info->block.levels,info->block.root,
+			(uchar*) 0));
+  }
+  info->block.levels=0;
+  info->del_chunk_count= info->chunk_count= 0;
+  info->del_link=0;
+  info->total_data_length= 0;
+}
+
+
+/**
+  Allocate or reallocate a chunkset in the dataspace
+
+  Attempts to allocate a new chunkset or change the size of an existing chunkset
+
+  @param  info            the hosting dataspace
+  @param  chunk_count     the number of chunks that we expect as the result
+  @param  existing_set    non-null value asks function to resize existing chunkset,
+                          return value would point to this set
+
+  @return  Pointer to the first chunk in the new or updated chunkset, or NULL if unsuccessful
+*/
+
+static uchar *hp_allocate_variable_chunkset(HP_DATASPACE *info,
+                                           uint chunk_count, uchar* existing_set)
+{
+  int alloc_count= chunk_count, i;
+  uchar *first_chunk= 0, *curr_chunk= 0, *prev_chunk= 0, *last_existing_chunk= 0;
+
+  assert(alloc_count);
+
+  if (existing_set)
+  {
+    first_chunk= existing_set;
+
+    curr_chunk= existing_set;
+    while (curr_chunk && alloc_count)
+    {
+      prev_chunk= curr_chunk;
+      curr_chunk= *((uchar**)(curr_chunk + info->offset_link));
+      alloc_count--;
+    }
+
+    if (!alloc_count)
+    {
+      if (curr_chunk)
+      {
+        /* We came through all chunks and there is more left, let's truncate the list */
+        *((uchar**)(prev_chunk + info->offset_link)) = NULL;
+        hp_free_chunks(info, curr_chunk);
+      }
+
+      return first_chunk;
+    }
+
+    last_existing_chunk = prev_chunk;
+  }
+
+  /* We can reach this point only if we're allocating new chunkset or more chunks in existing set */
+
+  for (i=0; i<alloc_count; i++)
+  {
+      curr_chunk= hp_allocate_one_chunk(info);
+      if (!curr_chunk)
+      {
+        /* no space in the current block */
+
+        if (last_existing_chunk)
+        {
+          /* Truncate whatever was added at the end of the existing chunkset */
+          prev_chunk= last_existing_chunk;
+          curr_chunk= *((uchar**)(prev_chunk + info->offset_link));
+          *((uchar**)(prev_chunk + info->offset_link)) = NULL;
+          hp_free_chunks(info, curr_chunk);
+        }
+        else if (first_chunk)
+        {
+          /* free any chunks previously allocated */
+          hp_free_chunks(info, first_chunk);
+        }
+
+        return NULL;
+      }
+
+      /* mark as if this chunk is last in the chunkset */
+      *((uchar**) (curr_chunk + info->offset_link))= 0;
+
+      if (prev_chunk)
+      {
+        /* tie them into a linked list */
+        *((uchar**) (prev_chunk + info->offset_link))= curr_chunk;
+        curr_chunk[info->offset_status]= CHUNK_STATUS_LINKED;			/* Record linked from active */
+      }
+      else
+      {
+        curr_chunk[info->offset_status]= CHUNK_STATUS_ACTIVE;			  /* Record active */
+      }
+
+      if (!first_chunk)
+      {
+        first_chunk= curr_chunk;
+      }
+
+      prev_chunk= curr_chunk;
+  }
+
+  return first_chunk;
+}
+
+
+/**
+  Allocate a new chunkset in the dataspace
+
+  Attempts to allocate a new chunkset
+
+  @param  info            the hosting dataspace
+  @param  chunk_count     the number of chunks that we expect as the result
+
+  @return  Pointer to the first chunk in the new or updated chunkset, or NULL if unsuccessful
+*/
+
+uchar *hp_allocate_chunkset(HP_DATASPACE *info, uint chunk_count)
+{
+  uchar* result;
+
+
+  if (info->is_variable_size)
+  {
+    result = hp_allocate_variable_chunkset(info, chunk_count, NULL);
+  }
+  else
+  {
+    result= hp_allocate_one_chunk(info);
+    if (result)
+    {
+      result[info->offset_status]= CHUNK_STATUS_ACTIVE;
+    }
+
+    return(result);
+  }
+
+  return(result);
+}
+
+
+/**
+  Reallocate an existing chunkset in the dataspace
+
+  Attempts to change the size of an existing chunkset
+
+  @param  info            the hosting dataspace
+  @param  chunk_count     the number of chunks that we expect as the result
+  @param  pos             pointer to the existing chunkset
+
+  @return  Error code or zero if successful
+*/
+
+int hp_reallocate_chunkset(HP_DATASPACE *info, uint chunk_count, uchar* pos)
+{
+
+  if (!info->is_variable_size)
+  {
+    /* Update should never change chunk_count in fixed-size mode */
+    my_errno=HA_ERR_WRONG_COMMAND;
+    return my_errno;
+  }
+
+  /* Reallocate never moves the first chunk */
+  if (!hp_allocate_variable_chunkset(info, chunk_count, pos))
+    return(my_errno);
+
+  return(0);
+}
+
+
+/**
+  Allocate a single chunk in the dataspace
+
+  Attempts to allocate a new chunk or reuse one from deleted list
+
+  @param  info            the hosting dataspace
+
+  @return  Pointer to the chunk, or NULL if unsuccessful
+*/
+
+static uchar *hp_allocate_one_chunk(HP_DATASPACE *info)
+{
+  uchar* curr_chunk;
+  size_t length, block_pos;
+
+  if (info->del_link)
+  {
+    curr_chunk=info->del_link;
+    info->del_link= *((uchar**) curr_chunk);
+    info->del_chunk_count--;
+
+    return curr_chunk;
+  }
+
+  block_pos= (info->chunk_count % info->block.records_in_block);
+  if (!block_pos)
+  {
+    if (hp_get_new_block(&info->block,&length))
+    {
+      /* no space in the current block */
+      return NULL;
+    }
+
+    info->total_data_length+= length;
+  }
+
+  info->chunk_count++;
+  curr_chunk= ((uchar*) info->block.level_info[0].last_blocks +
+    block_pos * info->block.recbuffer);
+
+
+  return curr_chunk;
+}
+
+
+/**
+  Free a list of chunks
+
+  Reclaims all chunks linked by the pointer,
+  which could be the whole chunkset or a part of an existing chunkset
+
+  @param  info            the hosting dataspace
+  @param  pos             pointer to the head of the chunkset
+*/
+
+void hp_free_chunks(HP_DATASPACE *info, uchar *pos)
+{
+  uchar* curr_chunk= pos;
+
+  while (curr_chunk) {
+    info->del_chunk_count++;
+    *((uchar**) curr_chunk)= info->del_link;
+    info->del_link= curr_chunk;
+
+    curr_chunk[info->offset_status]= CHUNK_STATUS_DELETED;
+
+
+    if (!info->is_variable_size)
+    {
+      break;
+    }
+
+    /* Delete next chunk in this chunkset */
+    curr_chunk= *((uchar**)(curr_chunk + info->offset_link));
+  }
+}

=== modified file 'storage/heap/hp_info.c'
--- storage/heap/hp_info.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_info.c	2008-08-05 05:01:04 +0000
@@ -30,9 +30,21 @@
 int heap_info(register HP_INFO *info, register HEAPINFO *x, int flag )
 {
   x->records         = info->s->records;
-  x->deleted         = info->s->deleted;
-  x->reclength       = info->s->reclength;
-  x->data_length     = info->s->data_length;
+  x->deleted         = info->s->recordspace.del_chunk_count;
+
+  if (info->s->recordspace.is_variable_size)
+  {
+    if (info->s->records)
+      x->reclength   = (uint)(info->s->recordspace.total_data_length / (uint64_t)info->s->records);
+    else
+      x->reclength   = info->s->recordspace.chunk_length;
+  }
+  else
+  {
+    x->reclength     = info->s->recordspace.chunk_dataspace_length;
+  }
+
+  x->data_length     = info->s->recordspace.total_data_length;
   x->index_length    = info->s->index_length;
   x->max_records     = info->s->max_records;
   x->errkey          = info->errkey;

=== added file 'storage/heap/hp_record.c'
--- storage/heap/hp_record.c	1970-01-01 00:00:00 +0000
+++ storage/heap/hp_record.c	2008-08-05 05:01:04 +0000
@@ -0,0 +1,369 @@
+/* Copyright (C) 2000-2002 MySQL AB
+   Copyright (C) 2008 eBay, Inc
+
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+
+/* Implements various base record-related functions, such as encode and decode into chunks */
+
+#include "heapdef.h"
+#include <libdrizzle/drizzle_com.h>
+
+
+/**
+  Calculate size of the record for the purpose of storing in chunks
+
+  Walk through the fields of the record and calculates the exact space
+  needed in chunks as well the the total chunk count
+
+  @param       info         the hosting table
+  @param       record       the record in standard unpacked format
+  @param[out]  chunk_count  the number of chunks needed for this record
+
+  @return The size of the required storage in bytes
+*/
+
+uint hp_get_encoded_data_length(HP_SHARE *info, const uchar *record, uint *chunk_count)
+{
+  uint i, dst_offset;
+
+  dst_offset= info->fixed_data_length;
+
+  if (!info->recordspace.is_variable_size)
+  {
+    /* Nothing more to copy */
+    *chunk_count= 1;
+    return dst_offset;
+  }
+
+  for (i= info->fixed_column_count; i < info->column_count; i++)
+  {
+    uint src_offset, length;
+
+    HP_COLUMNDEF* column= info->column_defs + i;
+
+    if (column->null_bit)
+    {
+      if (record[column->null_pos] & column->null_bit)
+      {
+        /* Skip all NULL values */
+        continue;
+      }
+    }
+
+    src_offset= column->offset;
+    if (column->type == DRIZZLE_TYPE_VARCHAR)
+    {
+      uint pack_length;
+
+      /* >= 5.0.3 true VARCHAR */
+
+      pack_length= column->length_bytes;
+      length= pack_length + (pack_length == 1 ?
+        (uint) *(uchar*) (record + src_offset) : uint2korr(record + src_offset));
+    }
+    else
+    {
+      length= column->length;
+    }
+
+    dst_offset+= length;
+  }
+
+  *chunk_count= get_chunk_count(&info->recordspace, dst_offset);
+
+  return dst_offset;
+}
+
+
+/*static void dump_chunk(HP_SHARE *info, const uchar* curr_chunk)
+{
+  uint i;
+  fprintf(stdout, "Chunk dump at 0x%lx: ", (long)curr_chunk);
+  for (i= 0; i < info->chunk_dataspace_length; i++)
+  {
+    uint b= *((uchar*)(curr_chunk + i));
+    if (b < 0x10)
+    {
+      fprintf(stdout, "0");
+    }
+    fprintf(stdout, "%lx ", (long)b);
+  }
+  fprintf(stdout, ". Next = 0x%lx, Status = %d\n",
+    (long)(*((uchar**) (curr_chunk + info->offset_link))),
+    (uint)(*((uchar*) (curr_chunk + info->offset_status))));
+}*/
+
+
+/**
+  Encodes or compares record
+
+  Copies data from original unpacked record into the preallocated chunkset,
+  or performs data comparison
+
+  @param  info         the hosting table
+  @param  record       the record in standard unpacked format
+  @param  pos          the target chunkset
+  @param  is_compare   flag indicating whether we should compare data or store it
+
+  @return  Status of comparison
+    @retval  non-zero  if comparison fond data differences
+    @retval  zero      otherwise
+*/
+
+uint hp_process_record_data_to_chunkset(HP_SHARE *info, const uchar *record,
+                                      uchar *pos, uint is_compare)
+{
+  uint i, dst_offset;
+  uchar* curr_chunk= pos;
+
+  if (is_compare)
+  {
+    if (memcmp(curr_chunk, record, (size_t) info->fixed_data_length))
+    {
+      return 1;
+    }
+  }
+  else
+  {
+    memcpy(curr_chunk, record, (size_t) info->fixed_data_length);
+  }
+
+  if (!info->recordspace.is_variable_size)
+  {
+    /* Nothing more to copy */
+    return 0;
+  }
+
+  dst_offset= info->fixed_data_length;
+
+  for (i= info->fixed_column_count; i < info->column_count; i++)
+  {
+    uint src_offset, length;
+
+    HP_COLUMNDEF* column= info->column_defs + i;
+
+    if (column->null_bit)
+    {
+      if (record[column->null_pos] & column->null_bit)
+      {
+        /* Skip all NULL values */
+        continue;
+      }
+    }
+
+    src_offset= column->offset;
+    if (column->type == DRIZZLE_TYPE_VARCHAR)
+    {
+      uint pack_length;
+
+      /* >= 5.0.3 true VARCHAR */
+
+      /* Make sure to copy length indicator and actuals string bytes */
+      pack_length= column->length_bytes;
+      length= pack_length + (pack_length == 1 ?
+        (uint) *(uchar*) (record + src_offset) : uint2korr(record + src_offset));
+    }
+    else
+    {
+      length= column->length;
+    }
+
+    while (length > 0)
+    {
+      uint to_copy;
+
+      to_copy= info->recordspace.chunk_dataspace_length - dst_offset;
+      if (to_copy == 0)
+      {
+        /* Jump to the next chunk */
+        /*dump_chunk(info, curr_chunk);*/
+        curr_chunk= *((uchar**) (curr_chunk + info->recordspace.offset_link));
+        dst_offset= 0;
+        continue;
+      }
+
+      to_copy= min(length, to_copy);
+
+      if (is_compare)
+      {
+        if (memcmp(curr_chunk + dst_offset, record + src_offset, (size_t) to_copy))
+        {
+          return 1;
+        }
+      }
+      else
+      {
+        memcpy(curr_chunk + dst_offset, record + src_offset, (size_t) to_copy);
+      }
+
+      src_offset+= to_copy;
+      dst_offset+= to_copy;
+      length-= to_copy;
+    }
+  }
+
+  /*dump_chunk(info, curr_chunk);*/
+  return 0;
+}
+
+
+/**
+  Stores record in the heap table chunks
+
+  Copies data from original unpacked record into the preallocated chunkset
+
+  @param  info         the hosting table
+  @param  record       the record in standard unpacked format
+  @param  pos          the target chunkset
+*/
+
+void hp_copy_record_data_to_chunkset(HP_SHARE *info, const uchar *record, uchar *pos)
+{
+
+  hp_process_record_data_to_chunkset(info, record, pos, 0);
+
+  return;
+}
+
+
+/*
+  Macro to switch curr_chunk to the next chunk in the chunkset and reset src_offset
+*/
+#define SWITCH_TO_NEXT_CHUNK_FOR_READ(info, curr_chunk, src_offset) \
+      { \
+        curr_chunk= *((uchar**) (curr_chunk + info->recordspace.offset_link)); \
+        src_offset= 0; \
+        /*dump_chunk(info, curr_chunk);*/ \
+      }
+
+
+/**
+  Copies record data from storage to unpacked record format
+
+  Copies data from chunkset into its original unpacked record
+
+  @param       info         the hosting table
+  @param[out]  record       the target record in standard unpacked format
+  @param       pos          the source chunkset
+*/
+
+void hp_extract_record(HP_SHARE *info, uchar *record, const uchar *pos)
+{
+  uint i, src_offset;
+  const uchar* curr_chunk= pos;
+
+
+  /*if (info->is_variable_size)
+  {
+    dump_chunk(info, curr_chunk);
+  }*/
+
+  memcpy(record, curr_chunk, (size_t) info->fixed_data_length);
+
+  if (!info->recordspace.is_variable_size)
+  {
+    /* Nothing more to copy */
+    return;
+  }
+
+  src_offset= info->fixed_data_length;
+
+  for (i= info->fixed_column_count; i < info->column_count; i++)
+  {
+    uint dst_offset, length, is_null = 0;
+
+    HP_COLUMNDEF* column= info->column_defs + i;
+
+    if (column->null_bit)
+    {
+      if (record[column->null_pos] & column->null_bit)
+      {
+        is_null = 1;
+      }
+    }
+
+    dst_offset= column->offset;
+    if (column->type == DRIZZLE_TYPE_VARCHAR)
+    {
+      uint pack_length, byte1, byte2;
+
+      /* >= 5.0.3 true VARCHAR */
+
+      if (is_null)
+      {
+        /* TODO: is memset really needed? */
+        memset(record + column->offset, 0, column->length);
+        continue;
+      }
+
+      pack_length= column->length_bytes;
+
+      if (src_offset == info->recordspace.chunk_dataspace_length)
+      {
+        SWITCH_TO_NEXT_CHUNK_FOR_READ(info, curr_chunk, src_offset);
+      }
+      byte1= *(uchar*) (curr_chunk + src_offset++);
+      *(record + dst_offset++)= byte1;
+
+      if (pack_length == 1)
+      {
+        length= byte1;
+      }
+      else
+      {
+        if (src_offset == info->recordspace.chunk_dataspace_length)
+        {
+          SWITCH_TO_NEXT_CHUNK_FOR_READ(info, curr_chunk, src_offset);
+        }
+        byte2= *(uchar*) (curr_chunk + src_offset++);
+        *(record + dst_offset++)= byte2;
+
+        /* We copy byte-by-byte and then use uint2korr to combine bytes in the right order */
+        length= uint2korr(record + dst_offset - 2);
+      }
+    }
+    else
+    {
+      if (is_null)
+      {
+        /* TODO: is memset really needed? */
+        memset(record + column->offset, 0, column->length);
+        continue;
+      }
+
+      length= column->length;
+    }
+
+    while (length > 0)
+    {
+      uint to_copy;
+
+      to_copy= info->recordspace.chunk_dataspace_length - src_offset;
+      if (to_copy == 0)
+      {
+        SWITCH_TO_NEXT_CHUNK_FOR_READ(info, curr_chunk, src_offset);
+        to_copy= info->recordspace.chunk_dataspace_length;
+      }
+
+      to_copy= min(length, to_copy);
+
+      memcpy(record + dst_offset, curr_chunk + src_offset, (size_t) to_copy);
+      src_offset+= to_copy;
+      dst_offset+= to_copy;
+      length-= to_copy;
+    }
+  }
+
+  return;
+}

=== added file 'storage/heap/hp_rectest.c'
--- storage/heap/hp_rectest.c	1970-01-01 00:00:00 +0000
+++ storage/heap/hp_rectest.c	2008-08-05 05:01:04 +0000
@@ -0,0 +1,29 @@
+/* Copyright (C) 2000-2002 MySQL AB
+
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+
+/* Test if a record has changed since last read */
+
+#include "heapdef.h"
+
+int hp_rectest(register HP_INFO *info, register const uchar *old_record)
+{
+
+  if (hp_process_record_data_to_chunkset(info->s, old_record, info->current_ptr, 1))
+  {
+    return((my_errno=HA_ERR_RECORD_CHANGED)); /* Record have changed */
+  }
+
+  return(0);
+} /* _heap_rectest */

=== modified file 'storage/heap/hp_rfirst.c'
--- storage/heap/hp_rfirst.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_rfirst.c	2008-08-05 05:01:04 +0000
@@ -33,7 +33,7 @@
       memcpy(&pos, pos + (*keyinfo->get_key_length)(keyinfo, pos), 
 	     sizeof(uchar*));
       info->current_ptr = pos;
-      memcpy(record, pos, (size_t)share->reclength);
+      hp_extract_record(share, record, pos);
       /*
         If we're performing index_first on a table that was taken from
         table cache, info->lastkey_len is initialized to previous query.

=== modified file 'storage/heap/hp_rkey.c'
--- storage/heap/hp_rkey.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_rkey.c	2008-08-05 05:01:04 +0000
@@ -64,7 +64,7 @@
     if (!(keyinfo->flag & HA_NOSAME))
       memcpy(info->lastkey, key, (size_t) keyinfo->length);
   }
-  memcpy(record, pos, (size_t) share->reclength);
+  hp_extract_record(share, record, pos);
   info->update= HA_STATE_AKTIV;
   return(0);
 }

=== modified file 'storage/heap/hp_rlast.c'
--- storage/heap/hp_rlast.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_rlast.c	2008-08-05 05:01:04 +0000
@@ -34,7 +34,7 @@
       memcpy(&pos, pos + (*keyinfo->get_key_length)(keyinfo, pos), 
 	     sizeof(uchar*));
       info->current_ptr = pos;
-      memcpy(record, pos, (size_t)share->reclength);
+      hp_extract_record(share, record, pos);
       info->update = HA_STATE_AKTIV;
     }
     else

=== modified file 'storage/heap/hp_rnext.c'
--- storage/heap/hp_rnext.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_rnext.c	2008-08-05 05:01:04 +0000
@@ -108,7 +108,7 @@
       my_errno=HA_ERR_END_OF_FILE;
     return(my_errno);
   }
-  memcpy(record,pos,(size_t) share->reclength);
+  hp_extract_record(share, record, pos);
   info->update=HA_STATE_AKTIV | HA_STATE_NEXT_FOUND;
   return(0);
 }

=== modified file 'storage/heap/hp_rprev.c'
--- storage/heap/hp_rprev.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_rprev.c	2008-08-05 05:01:04 +0000
@@ -76,7 +76,7 @@
       my_errno=HA_ERR_END_OF_FILE;
     return(my_errno);
   }
-  memcpy(record,pos,(size_t) share->reclength);
+  hp_extract_record(share, record, pos);
   info->update=HA_STATE_AKTIV | HA_STATE_PREV_FOUND;
   return(0);
 }

=== modified file 'storage/heap/hp_rrnd.c'
--- storage/heap/hp_rrnd.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_rrnd.c	2008-08-05 05:01:04 +0000
@@ -34,13 +34,14 @@
     info->update= 0;
     return(my_errno= HA_ERR_END_OF_FILE);
   }
-  if (!info->current_ptr[share->reclength])
+  if (get_chunk_status(&share->recordspace, info->current_ptr) != CHUNK_STATUS_ACTIVE)
   {
+    /* treat deleted and linked chunks as deleted */
     info->update= HA_STATE_PREV_FOUND | HA_STATE_NEXT_FOUND;
     return(my_errno=HA_ERR_RECORD_DELETED);
   }
   info->update=HA_STATE_PREV_FOUND | HA_STATE_NEXT_FOUND | HA_STATE_AKTIV;
-  memcpy(record,info->current_ptr,(size_t) share->reclength);
+  hp_extract_record(share, record, info->current_ptr);
   info->current_hash_ptr=0;			/* Can't use rnext */
   return(0);
 } /* heap_rrnd */
@@ -59,23 +60,23 @@
 int heap_rrnd_old(register HP_INFO *info, uchar *record, ulong pos)
 {
   HP_SHARE *share=info->s;
-
+asdfasdf;
   info->lastinx= -1;
   if (pos == (ulong) -1)
   {
     pos= ++info->current_record;
     if (pos % share->block.records_in_block &&	/* Quick next record */
-	pos < share->records+share->deleted &&
-	(info->update & HA_STATE_PREV_FOUND))
+      pos < share->used_chunk_count+share->deleted_chunk_count &&
+	    (info->update & HA_STATE_PREV_FOUND))
     {
-      info->current_ptr+=share->block.recbuffer;
+      info->current_ptr+=share->block.recbufferlen;
       goto end;
     }
   }
   else
     info->current_record=pos;
 
-  if (pos >= share->records+share->deleted)
+  if (pos >= share->used_chunk_count+share->deleted_chunk_count)
   {
     info->update= 0;
     return(my_errno= HA_ERR_END_OF_FILE);

=== modified file 'storage/heap/hp_rsame.c'
--- storage/heap/hp_rsame.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_rsame.c	2008-08-05 05:01:04 +0000
@@ -30,7 +30,7 @@
   HP_SHARE *share=info->s;
 
   test_active(info);
-  if (info->current_ptr[share->reclength])
+  if (get_chunk_status(&share->recordspace, info->current_ptr) == CHUNK_STATUS_ACTIVE)
   {
     if (inx < -1 || inx >= (int) share->keys)
     {
@@ -46,7 +46,7 @@
 	return(my_errno);
       }
     }
-    memcpy(record,info->current_ptr,(size_t) share->reclength);
+    hp_extract_record(share, record, info->current_ptr);
     return(0);
   }
   info->update=0;

=== modified file 'storage/heap/hp_scan.c'
--- storage/heap/hp_scan.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_scan.c	2008-08-05 05:01:04 +0000
@@ -41,14 +41,14 @@
   pos= ++info->current_record;
   if (pos < info->next_block)
   {
-    info->current_ptr+=share->block.recbuffer;
+    info->current_ptr+=share->recordspace.block.recbuffer;
   }
   else
   {
-    info->next_block+=share->block.records_in_block;
-    if (info->next_block >= share->records+share->deleted)
+    info->next_block+=share->recordspace.block.records_in_block;
+    if (info->next_block >= share->recordspace.chunk_count)
     {
-      info->next_block= share->records+share->deleted;
+      info->next_block= share->recordspace.chunk_count;
       if (pos >= info->next_block)
       {
 	info->update= 0;
@@ -57,13 +57,13 @@
     }
     hp_find_record(info, pos);
   }
-  if (!info->current_ptr[share->reclength])
+  if (get_chunk_status(&share->recordspace, info->current_ptr) != CHUNK_STATUS_ACTIVE)
   {
     info->update= HA_STATE_PREV_FOUND | HA_STATE_NEXT_FOUND;
     return(my_errno=HA_ERR_RECORD_DELETED);
   }
   info->update= HA_STATE_PREV_FOUND | HA_STATE_NEXT_FOUND | HA_STATE_AKTIV;
-  memcpy(record,info->current_ptr,(size_t) share->reclength);
+  hp_extract_record(share, record, info->current_ptr);
   info->current_hash_ptr=0;			/* Can't use read_next */
   return(0);
 } /* heap_scan */

=== modified file 'storage/heap/hp_update.c'
--- storage/heap/hp_update.c	2008-07-17 16:34:04 +0000
+++ storage/heap/hp_update.c	2008-08-05 05:01:04 +0000
@@ -17,39 +17,56 @@
 
 #include "heapdef.h"
 
-int heap_update(HP_INFO *info, const uchar *old, const uchar *heap_new)
+int heap_update(HP_INFO *info, const uchar *old_record, const uchar *new_record)
 {
   HP_KEYDEF *keydef, *end, *p_lastinx;
   uchar *pos;
   my_bool auto_key_changed= 0;
   HP_SHARE *share= info->s;
+  uint old_length, new_length;
+  uint old_chunk_count, new_chunk_count;
 
   test_active(info);
   pos=info->current_ptr;
 
-  if (info->opt_flag & READ_CHECK_USED)
+  if (info->opt_flag & READ_CHECK_USED && hp_rectest(info,old_record))
     return(my_errno);				/* Record changed */
+
+  old_length = hp_get_encoded_data_length(share, old_record, &old_chunk_count);
+  new_length = hp_get_encoded_data_length(share, new_record, &new_chunk_count);
+
+  if (new_chunk_count > old_chunk_count) {
+    /* extend the old chunkset size as necessary, but do not shrink yet */
+    if (hp_reallocate_chunkset(&share->recordspace, new_chunk_count, pos)) {
+      return(my_errno);                          /* Out of memory or table space */
+    }
+  }
+ 
   if (--(share->records) < share->blength >> 1) share->blength>>= 1;
   share->changed=1;
 
   p_lastinx= share->keydef + info->lastinx;
   for (keydef= share->keydef, end= keydef + share->keys; keydef < end; keydef++)
   {
-    if (hp_rec_key_cmp(keydef, old, heap_new, 0))
+    if (hp_rec_key_cmp(keydef, old_record, new_record, 0))
     {
-      if ((*keydef->delete_key)(info, keydef, old, pos, keydef == p_lastinx) ||
-          (*keydef->write_key)(info, keydef, heap_new, pos))
+      if ((*keydef->delete_key)(info, keydef, old_record, pos, keydef == p_lastinx) ||
+          (*keydef->write_key)(info, keydef, new_record, pos))
         goto err;
       if (share->auto_key == (uint) (keydef - share->keydef + 1))
         auto_key_changed= 1;
     }
   }
-
-  memcpy(pos,heap_new,(size_t) share->reclength);
+  hp_copy_record_data_to_chunkset(share, new_record, pos);
   if (++(share->records) == share->blength) share->blength+= share->blength;
 
+  if (new_chunk_count < old_chunk_count) {
+    /* Shrink the chunkset to its new size */
+    hp_reallocate_chunkset(&share->recordspace, new_chunk_count, pos);
+  }
+
   if (auto_key_changed)
-    heap_update_auto_increment(info, heap_new);
+    heap_update_auto_increment(info, new_record);
   return(0);
 
  err:
@@ -59,7 +76,7 @@
     if (keydef->algorithm == HA_KEY_ALG_BTREE)
     {
       /* we don't need to delete non-inserted key from rb-tree */
-      if ((*keydef->write_key)(info, keydef, old, pos))
+      if ((*keydef->write_key)(info, keydef, old_record, pos))
       {
         if (++(share->records) == share->blength)
 	  share->blength+= share->blength;
@@ -69,16 +86,23 @@
     }
     while (keydef >= share->keydef)
     {
-      if (hp_rec_key_cmp(keydef, old, heap_new, 0))
+      if (hp_rec_key_cmp(keydef, old_record, new_record, 0))
       {
-	if ((*keydef->delete_key)(info, keydef, heap_new, pos, 0) ||
-	    (*keydef->write_key)(info, keydef, old, pos))
+	if ((*keydef->delete_key)(info, keydef, new_record, pos, 0) ||
+	    (*keydef->write_key)(info, keydef, old_record, pos))
 	  break;
       }
       keydef--;
     }
   }
+
   if (++(share->records) == share->blength)
     share->blength+= share->blength;
+
+  if (new_chunk_count > old_chunk_count) {
+    /* Shrink the chunkset to its original size */
+    hp_reallocate_chunkset(&share->recordspace, old_chunk_count, pos);
+  }
+
   return(my_errno);
 } /* heap_update */

=== modified file 'storage/heap/hp_write.c'
--- storage/heap/hp_write.c	2008-07-16 16:14:22 +0000
+++ storage/heap/hp_write.c	2008-08-05 05:01:04 +0000
@@ -25,7 +25,6 @@
 #define HIGHFIND 4
 #define HIGHUSED 8
 
-static uchar *next_free_record_pos(HP_SHARE *info);
 static HASH_INFO *hp_find_free_hash(HP_SHARE *info, HP_BLOCK *block,
 				     ulong records);
 
@@ -34,7 +33,17 @@
   HP_KEYDEF *keydef, *end;
   uchar *pos;
   HP_SHARE *share=info->s;
-  if (!(pos=next_free_record_pos(share)))
+  uint rec_length, chunk_count;
+
+  if ((share->records >= share->max_records && share->max_records) ||
+    (share->recordspace.total_data_length + share->index_length >= share->max_table_size))
+  {
+    return(my_errno=HA_ERR_RECORD_FILE_FULL);
+  }
+
+  rec_length = hp_get_encoded_data_length(share, record, &chunk_count);
+
+  if (!(pos=hp_allocate_chunkset(&share->recordspace, chunk_count)))
     return(my_errno);
   share->changed=1;
 
@@ -45,10 +54,11 @@
       goto err;
   }
 
-  memcpy(pos,record,(size_t) share->reclength);
-  pos[share->reclength]=1;		/* Mark record as not deleted */
+  hp_copy_record_data_to_chunkset(share, record, pos);
+
   if (++share->records == share->blength)
     share->blength+= share->blength;
+
   info->current_ptr=pos;
   info->current_hash_ptr=0;
   info->update|=HA_STATE_AKTIV;
@@ -75,10 +85,7 @@
     keydef--;
   } 
 
-  share->deleted++;
-  *((uchar**) pos)=share->del_link;
-  share->del_link=pos;
-  pos[share->reclength]=0;			/* Record deleted */
+  hp_free_chunks(&share->recordspace, pos);
 
   return(my_errno);
 } /* heap_write */
@@ -116,38 +123,6 @@
   return 0;
 }
 
-	/* Find where to place new record */
-
-static uchar *next_free_record_pos(HP_SHARE *info)
-{
-  int block_pos;
-  uchar *pos;
-  size_t length;
-
-  if (info->del_link)
-  {
-    pos=info->del_link;
-    info->del_link= *((uchar**) pos);
-    info->deleted--;
-    return(pos);
-  }
-  if (!(block_pos=(info->records % info->block.records_in_block)))
-  {
-    if ((info->records > info->max_records && info->max_records) ||
-        (info->data_length + info->index_length >= info->max_table_size))
-    {
-      my_errno=HA_ERR_RECORD_FILE_FULL;
-      return(NULL);
-    }
-    if (hp_get_new_block(&info->block,&length))
-      return(NULL);
-    info->data_length+=length;
-  }
-  return((uchar*) info->block.level_info[0].last_blocks+
-	      block_pos*info->block.recbuffer);
-}
-
-
 /*
   Write a hash-key to the hash-index
   SYNOPSIS

=== modified file 'tests/Makefile.am'
--- tests/Makefile.am	2008-08-01 04:36:33 +0000
+++ tests/Makefile.am	2008-08-05 06:19:05 +0000
@@ -141,6 +141,9 @@
 	  join_outer \
 	  distinct \
 	  type_newdecimal \
+	  heap_auto_increment \
+	  heap_btree \
+	  heap_var \
 	  join \
 	  subselect \
 	  subselect3 \
@@ -151,4 +154,5 @@
 	  information_schema \
 	  mysqlslap \
 	  lock \
-	  lock_tables_lost_commit
+	  lock_tables_lost_commit \
+	  heap_var

=== modified file 'tests/r/heap.result'
--- tests/r/heap.result	2008-06-25 05:29:13 +0000
+++ tests/r/heap.result	2008-08-05 06:19:05 +0000
@@ -169,7 +169,7 @@
 1	SIMPLE	t1	ALL	btn	NULL	NULL	NULL	14	Using where
 explain select * from t1 where btn="a" and new_col="a";
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
-1	SIMPLE	t1	ref	btn	btn	11	const,const	2	Using where
+1	SIMPLE	t1	ref	btn	btn	15	const,const	2	Using where
 drop table t1;
 CREATE TABLE t1 (
 a int default NULL,
@@ -218,7 +218,7 @@
 a
 DROP TABLE t1;
 CREATE TABLE `job_titles` (
-`job_title_id` int(6) unsigned NOT NULL default '0',
+`job_title_id` int unsigned NOT NULL default '0',
 `job_title` char(18) NOT NULL default '',
 PRIMARY KEY  (`job_title_id`),
 UNIQUE KEY `job_title_id` (`job_title_id`,`job_title`)
@@ -233,7 +233,7 @@
 a	B
 1	1
 DROP TABLE t1;
-CREATE TABLE t1 (pseudo char(35) PRIMARY KEY, date int(10) unsigned NOT NULL) ENGINE=HEAP;
+CREATE TABLE t1 (pseudo char(35) PRIMARY KEY, date int unsigned NOT NULL) ENGINE=HEAP;
 INSERT INTO t1 VALUES ('massecot',1101106491),('altec',1101106492),('stitch+',1101106304),('Seb Corgan',1101106305),('beerfilou',1101106263),('flaker',1101106529),('joce8',5),('M4vrick',1101106418),('gabay008',1101106525),('Vamp irX',1101106291),('ZoomZip',1101106546),('rip666',1101106502),('CBP ',1101106397),('guezpard',1101106496);
 DELETE FROM t1 WHERE date<1101106546;
 SELECT * FROM t1;
@@ -259,72 +259,70 @@
 Note	1265	Data truncated for column 'c' at row 1
 select concat('*',v,'*',c,'*',t,'*') from t1;
 concat('*',v,'*',c,'*',t,'*')
-*+ *+*+ *
-*+         *+*+                    *
+*+ *+ *+ *
+*+         *+         *+                    *
 show create table t1;
 Table	Create Table
-t1	CREATE TABLE `t1` (
-  `v` varchar(10) DEFAULT NULL,
-  `c` char(10) DEFAULT NULL,
-  `t` varchar(50) DEFAULT NULL
+t1	CREATE TABLE "t1" (
+  "v" varchar(10),
+  "c" varchar(10),
+  "t" varchar(50)
 ) ENGINE=MEMORY DEFAULT CHARSET=latin1
 create table t2 like t1;
 show create table t2;
 Table	Create Table
-t2	CREATE TABLE `t2` (
-  `v` varchar(10) DEFAULT NULL,
-  `c` char(10) DEFAULT NULL,
-  `t` varchar(50) DEFAULT NULL
+t2	CREATE TABLE "t2" (
+  "v" varchar(10),
+  "c" varchar(10),
+  "t" varchar(50)
 ) ENGINE=MEMORY DEFAULT CHARSET=latin1
 create table t3 select * from t1;
 show create table t3;
 Table	Create Table
-t3	CREATE TABLE `t3` (
-  `v` varchar(10) DEFAULT NULL,
-  `c` char(10) DEFAULT NULL,
-  `t` varchar(50) DEFAULT NULL
+t3	CREATE TABLE "t3" (
+  "v" varchar(10),
+  "c" varchar(10),
+  "t" varchar(50)
 ) ENGINE=MEMORY DEFAULT CHARSET=latin1
 alter table t1 modify c varchar(10);
 show create table t1;
 Table	Create Table
-t1	CREATE TABLE `t1` (
-  `v` varchar(10) DEFAULT NULL,
-  `c` varchar(10) DEFAULT NULL,
-  `t` varchar(50) DEFAULT NULL
+t1	CREATE TABLE "t1" (
+  "v" varchar(10),
+  "c" varchar(10),
+  "t" varchar(50)
 ) ENGINE=MEMORY DEFAULT CHARSET=latin1
 alter table t1 modify v char(10);
 show create table t1;
 Table	Create Table
-t1	CREATE TABLE `t1` (
-  `v` char(10) DEFAULT NULL,
-  `c` varchar(10) DEFAULT NULL,
-  `t` varchar(50) DEFAULT NULL
+t1	CREATE TABLE "t1" (
+  "v" varchar(10),
+  "c" varchar(10),
+  "t" varchar(50)
 ) ENGINE=MEMORY DEFAULT CHARSET=latin1
-alter table t1 modify t varchar(10);
-Warnings:
-Warning	1265	Data truncated for column 't' at row 2
+alter table t1 modify t varchar(50);
 show create table t1;
 Table	Create Table
-t1	CREATE TABLE `t1` (
-  `v` char(10) DEFAULT NULL,
-  `c` varchar(10) DEFAULT NULL,
-  `t` varchar(10) DEFAULT NULL
+t1	CREATE TABLE "t1" (
+  "v" varchar(10),
+  "c" varchar(10),
+  "t" varchar(50)
 ) ENGINE=MEMORY DEFAULT CHARSET=latin1
 select concat('*',v,'*',c,'*',t,'*') from t1;
 concat('*',v,'*',c,'*',t,'*')
-*+*+*+ *
-*+*+*+         *
+*+ *+ *+ *
+*+         *+         *+                    *
 drop table t1,t2,t3;
 create table t1 (v varchar(10), c char(10), t varchar(50), key(v), key(c), key(t(10)));
 show create table t1;
 Table	Create Table
-t1	CREATE TABLE `t1` (
-  `v` varchar(10) DEFAULT NULL,
-  `c` char(10) DEFAULT NULL,
-  `t` varchar(50) DEFAULT NULL,
-  KEY `v` (`v`),
-  KEY `c` (`c`),
-  KEY `t` (`t`(10))
+t1	CREATE TABLE "t1" (
+  "v" varchar(10),
+  "c" varchar(10),
+  "t" varchar(50),
+  KEY "v" ("v"),
+  KEY "c" ("c"),
+  KEY "t" ("t"(10))
 ) ENGINE=MEMORY DEFAULT CHARSET=latin1
 select count(*) from t1;
 count(*)
@@ -371,7 +369,7 @@
 1	SIMPLE	t1	ref	v	v	13	const	10	Using where
 explain select count(*) from t1 where c='a  ';
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
-1	SIMPLE	t1	ref	c	c	11	const	10	Using where
+1	SIMPLE	t1	ref	c	c	13	const	10	Using where
 explain select count(*) from t1 where t='a  ';
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
 1	SIMPLE	t1	ref	t	t	13	const	10	Using where
@@ -389,15 +387,15 @@
 select concat('*',v,'*',c,'*',t,'*') as qq from t1 where v='a' order by length(concat('*',v,'*',c,'*',t,'*'));
 qq
 *a*a*a*
-*a *a*a *
-*a  *a*a  *
-*a   *a*a   *
-*a    *a*a    *
-*a     *a*a     *
-*a      *a*a      *
-*a       *a*a       *
-*a        *a*a        *
-*a         *a*a         *
+*a *a *a *
+*a  *a  *a  *
+*a   *a   *a   *
+*a    *a    *a    *
+*a     *a     *a     *
+*a      *a      *a      *
+*a       *a       *a       *
+*a        *a        *a        *
+*a         *a         *a         *
 explain select * from t1 where v='a';
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
 1	SIMPLE	t1	ref	v	v	13	const	10	Using where
@@ -488,15 +486,15 @@
 select sql_big_result c,count(t) from t1 group by c limit 10;
 c	count(t)
 a	1
-a	10
-b	10
-c	10
-d	10
-e	10
-f	10
-g	10
-h	10
-i	10
+a         	10
+b   	10
+c      	10
+d  	10
+e     	10
+f        	10
+g  	10
+h     	10
+i 	10
 select t,count(*) from t1 group by t limit 10;
 t	count(*)
 a	1
@@ -537,7 +535,7 @@
 create table t1 (a char(10), unique (a));
 insert into t1 values ('a');
 insert into t1 values ('a ');
-ERROR 23000: Duplicate entry 'a' for key 'a'
+ERROR 23000: Duplicate entry 'a ' for key 'a'
 alter table t1 modify a varchar(10);
 insert into t1 values ('a '),('a  '),('a   '),('a         ');
 ERROR 23000: Duplicate entry 'a ' for key 'a'
@@ -553,13 +551,13 @@
 create table t1 (v varchar(10), c char(10), t varchar(50), key using btree (v), key using btree (c), key using btree (t(10)));
 show create table t1;
 Table	Create Table
-t1	CREATE TABLE `t1` (
-  `v` varchar(10) DEFAULT NULL,
-  `c` char(10) DEFAULT NULL,
-  `t` varchar(50) DEFAULT NULL,
-  KEY `v` (`v`) USING BTREE,
-  KEY `c` (`c`) USING BTREE,
-  KEY `t` (`t`(10)) USING BTREE
+t1	CREATE TABLE "t1" (
+  "v" varchar(10),
+  "c" varchar(10),
+  "t" varchar(50),
+  KEY "v" ("v") USING BTREE,
+  KEY "c" ("c") USING BTREE,
+  KEY "t" ("t"(10)) USING BTREE
 ) ENGINE=MEMORY DEFAULT CHARSET=latin1
 select count(*) from t1;
 count(*)
@@ -594,7 +592,7 @@
 1	SIMPLE	t1	ref	v	v	13	const	#	Using where
 explain select count(*) from t1 where c='a  ';
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
-1	SIMPLE	t1	ref	c	c	11	const	#	Using where
+1	SIMPLE	t1	ref	c	c	13	const	#	Using where
 explain select count(*) from t1 where t='a  ';
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
 1	SIMPLE	t1	ref	t	t	13	const	#	Using where
@@ -612,15 +610,15 @@
 select concat('*',v,'*',c,'*',t,'*') as qq from t1 where v='a' order by length(concat('*',v,'*',c,'*',t,'*'));
 qq
 *a*a*a*
-*a *a*a *
-*a  *a*a  *
-*a   *a*a   *
-*a    *a*a    *
-*a     *a*a     *
-*a      *a*a      *
-*a       *a*a       *
-*a        *a*a        *
-*a         *a*a         *
+*a *a *a *
+*a  *a  *a  *
+*a   *a   *a   *
+*a    *a    *a    *
+*a     *a     *a     *
+*a      *a      *a      *
+*a       *a       *a       *
+*a        *a        *a        *
+*a         *a         *a         *
 explain select * from t1 where v='a';
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
 1	SIMPLE	t1	ref	v	v	13	const	#	Using where
@@ -628,7 +626,7 @@
 create table t1 (a char(10), unique using btree (a)) engine=heap;
 insert into t1 values ('a');
 insert into t1 values ('a ');
-ERROR 23000: Duplicate entry 'a' for key 'a'
+ERROR 23000: Duplicate entry 'a ' for key 'a'
 alter table t1 modify a varchar(10);
 insert into t1 values ('a '),('a  '),('a   '),('a         ');
 ERROR 23000: Duplicate entry 'a ' for key 'a'
@@ -644,28 +642,28 @@
 create table t1 (v varchar(10), c char(10), t varchar(50), key(v(5)), key(c(5)), key(t(5)));
 show create table t1;
 Table	Create Table
-t1	CREATE TABLE `t1` (
-  `v` varchar(10) DEFAULT NULL,
-  `c` char(10) DEFAULT NULL,
-  `t` varchar(50) DEFAULT NULL,
-  KEY `v` (`v`(5)),
-  KEY `c` (`c`(5)),
-  KEY `t` (`t`(5))
+t1	CREATE TABLE "t1" (
+  "v" varchar(10),
+  "c" varchar(10),
+  "t" varchar(50),
+  KEY "v" ("v"(5)),
+  KEY "c" ("c"(5)),
+  KEY "t" ("t"(5))
 ) ENGINE=MEMORY DEFAULT CHARSET=latin1
 drop table t1;
 create table t1 (v varchar(65530), key(v(10)));
 show create table t1;
 Table	Create Table
-t1	CREATE TABLE `t1` (
-  `v` varchar(65530) DEFAULT NULL,
-  KEY `v` (`v`(10))
+t1	CREATE TABLE "t1" (
+  "v" varchar(65530),
+  KEY "v" ("v"(10))
 ) ENGINE=MEMORY DEFAULT CHARSET=latin1
 insert into t1 values(repeat('a',65530));
 select length(v) from t1 where v=repeat('a',65530);
 length(v)
 65530
 drop table t1;
-set storage_engine=MyISAM;
+set storage_engine=InnoDB;
 create table t1 (a bigint unsigned auto_increment primary key, b int,
 key (b, a)) engine=heap;
 insert t1 (b) values (1),(1),(1),(1),(1),(1),(1),(1);
@@ -716,7 +714,7 @@
 create table t2 (c varchar(10)) engine=memory;
 show table status like 't_';
 Name	Engine	Version	Row_format	Rows	Avg_row_length	Data_length	Max_data_length	Index_length	Data_free	Auto_increment	Create_time	Update_time	Check_time	Collation	Checksum	Create_options	Comment
-t1	MEMORY	10	Fixed	0	11	0	#	0	0	NULL	NULL	NULL	NULL	latin1_swedish_ci	NULL		
+t1	MEMORY	10	Fixed	0	12	0	#	0	0	NULL	NULL	NULL	NULL	latin1_swedish_ci	NULL		
 t2	MEMORY	10	Fixed	0	12	0	#	0	0	NULL	NULL	NULL	NULL	latin1_swedish_ci	NULL		
 drop table t1, t2;
 CREATE TABLE t1(a VARCHAR(1), b VARCHAR(2), c VARCHAR(256),

=== modified file 'tests/r/heap_btree.result'
--- tests/r/heap_btree.result	2008-06-25 05:29:13 +0000
+++ tests/r/heap_btree.result	2008-08-05 06:19:05 +0000
@@ -175,16 +175,16 @@
 insert into t1 values ("hello"),("hello"),("hello"),("hello"),("hello"),("a"),("b"),("c"),("d"),("e"),("f"),("g"),("h"),("i");
 explain select * from t1 where btn like "i%";
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
-1	SIMPLE	t1	range	btn	btn	10	NULL	1	Using where
+1	SIMPLE	t1	range	btn	btn	12	NULL	1	Using where
 explain select * from t1 where btn like "h%";
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
-1	SIMPLE	t1	range	btn	btn	10	NULL	#	Using where
+1	SIMPLE	t1	range	btn	btn	12	NULL	#	Using where
 explain select * from t1 where btn like "a%";
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
-1	SIMPLE	t1	range	btn	btn	10	NULL	1	Using where
+1	SIMPLE	t1	range	btn	btn	12	NULL	1	Using where
 explain select * from t1 where btn like "b%";
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
-1	SIMPLE	t1	range	btn	btn	10	NULL	1	Using where
+1	SIMPLE	t1	range	btn	btn	12	NULL	1	Using where
 select * from t1 where btn like "ff%";
 btn
 select * from t1 where btn like " %";
@@ -195,10 +195,10 @@
 update t1 set new_col=left(btn,1);
 explain select * from t1 where btn="a";
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
-1	SIMPLE	t1	ref	btn	btn	10	const	1	Using where
+1	SIMPLE	t1	ref	btn	btn	12	const	1	Using where
 explain select * from t1 where btn="a" and new_col="a";
 id	select_type	table	type	possible_keys	key	key_len	ref	rows	Extra
-1	SIMPLE	t1	ref	btn	btn	11	const,const	1	Using where
+1	SIMPLE	t1	ref	btn	btn	15	const,const	1	Using where
 drop table t1;
 CREATE TABLE t1 (
 a int default NULL,

=== added file 'tests/r/heap_var.result'
--- tests/r/heap_var.result	1970-01-01 00:00:00 +0000
+++ tests/r/heap_var.result	2008-08-05 05:01:04 +0000
@@ -0,0 +1,79 @@
+drop table if exists t1;
+set @@session.max_heap_table_size=16*1024*1024;
+create table t1 (a int not null, b varchar(400), c int, primary key (a), key (c)) engine=heap comment="testing heaps" block_size=128;
+ERROR 42000: Incorrect usage/placement of 'block_size'
+create table t1 (a int not null, b int, c varchar(400), primary key (a), key (b)) engine=heap comment="testing heaps" block_size=4;
+ERROR 42000: Incorrect usage/placement of 'block_size'
+create table t1 (a int not null, b int, c varchar(400), d varchar(400), primary key (a), key (b)) engine=heap comment="testing heaps" block_size=24;
+insert into t1 values (1,1,'012',NULL), (2,2,'0123456789',NULL), (3,3,'012345678901234567890123456789',NULL), (4,4,NULL,'0123456789012345678901234567890123456789012345678901234567890123456789');
+select * from t1;
+a	b	c	d
+1	1	012	NULL
+2	2	0123456789	NULL
+3	3	012345678901234567890123456789	NULL
+4	4	NULL	0123456789012345678901234567890123456789012345678901234567890123456789
+delete from t1 where a = 3;
+select * from t1;
+a	b	c	d
+1	1	012	NULL
+2	2	0123456789	NULL
+4	4	NULL	0123456789012345678901234567890123456789012345678901234567890123456789
+insert into t1 values (5,5,NULL,'0123'), (6,6,NULL,'0123');
+select * from t1;
+a	b	c	d
+1	1	012	NULL
+2	2	0123456789	NULL
+6	6	NULL	0123
+5	5	NULL	0123
+4	4	NULL	0123456789012345678901234567890123456789012345678901234567890123456789
+update t1 set c = '012345678901234567890123456789' where a = 2;
+select * from t1;
+a	b	c	d
+1	1	012	NULL
+2	2	012345678901234567890123456789	NULL
+6	6	NULL	0123
+5	5	NULL	0123
+4	4	NULL	0123456789012345678901234567890123456789012345678901234567890123456789
+update t1 set c = '0123456789' where a = 2;
+select * from t1;
+a	b	c	d
+1	1	012	NULL
+2	2	0123456789	NULL
+6	6	NULL	0123
+5	5	NULL	0123
+4	4	NULL	0123456789012345678901234567890123456789012345678901234567890123456789
+insert into t1 values (7,7,'0123',NULL), (8,8,'0123',NULL);
+select * from t1;
+a	b	c	d
+1	1	012	NULL
+2	2	0123456789	NULL
+6	6	NULL	0123
+5	5	NULL	0123
+4	4	NULL	0123456789012345678901234567890123456789012345678901234567890123456789
+7	7	0123	NULL
+8	8	0123	NULL
+alter table t1 block_size = 0;
+alter table t1 row_format = dynamic;
+alter table t1 block_size = 128, max_rows = 10001;
+select * from t1;
+a	b	c	d
+1	1	012	NULL
+2	2	0123456789	NULL
+6	6	NULL	0123
+5	5	NULL	0123
+4	4	NULL	0123456789012345678901234567890123456789012345678901234567890123456789
+7	7	0123	NULL
+8	8	0123	NULL
+delete from t1;
+select * from t1;
+a	b	c	d
+select count(*) from t1;
+count(*)
+10001
+insert into t1 values (100000,100000,NULL,'0123'), (100000,100000,NULL,'0123');
+ERROR HY000: The table 't1' is full
+select count(*) from t1;
+count(*)
+10001
+set @@session.max_heap_table_size=default;
+drop table t1;

=== modified file 'tests/t/archive.test'
--- tests/t/archive.test	2008-06-25 05:29:13 +0000
+++ tests/t/archive.test	2008-08-05 06:01:00 +0000
@@ -14,8 +14,8 @@
 SHOW VARIABLES LIKE "archive_%";
 
 CREATE TABLE t1 (
-  Period smallint(4) unsigned zerofill DEFAULT '0000' NOT NULL,
-  Varor_period smallint(4) unsigned DEFAULT '0' NOT NULL
+  Period smallint unsigned DEFAULT '0000' NOT NULL,
+  Varor_period smallint unsigned DEFAULT '0' NOT NULL
 ) ENGINE=archive;
 
 INSERT INTO t1 VALUES (9410,9412);
@@ -30,8 +30,8 @@
 
 CREATE TABLE t2 (
   auto int,
-  fld1 int(6) unsigned zerofill DEFAULT '000000' NOT NULL,
-  companynr tinyint(2) unsigned zerofill DEFAULT '00' NOT NULL,
+  fld1 int unsigned DEFAULT '000000' NOT NULL,
+  companynr tinyint unsigned DEFAULT '00' NOT NULL,
   fld3 char(30) DEFAULT '' NOT NULL,
   fld4 char(35) DEFAULT '' NOT NULL,
   fld5 char(35) DEFAULT '' NOT NULL,
@@ -1378,7 +1378,7 @@
 # First the simple stuff
 
 CREATE TABLE `t5` (
-`a` int(11) NOT NULL auto_increment,
+`a` int NOT NULL auto_increment,
 b char(12),
 PRIMARY KEY  (`a`)
 )  DEFAULT CHARSET=latin1;
@@ -1403,7 +1403,7 @@
 DROP TABLE t5;
 
 CREATE TABLE `t5` (
-`a` int(11) NOT NULL auto_increment,
+`a` int NOT NULL auto_increment,
 b char(12),
 KEY  (`a`)
 )  DEFAULT CHARSET=latin1 AUTO_INCREMENT=5;
@@ -1429,7 +1429,7 @@
 DROP TABLE t5;
 
 CREATE TABLE `t5` (
-`a` int(11) NOT NULL auto_increment,
+`a` int NOT NULL auto_increment,
 b blob(12),
 KEY  (`a`)
 )  DEFAULT CHARSET=latin1;
@@ -1456,7 +1456,7 @@
 DROP TABLE t5;
 
 CREATE TABLE `t5` (
-`a` int(11) NOT NULL auto_increment,
+`a` int NOT NULL auto_increment,
 b blob(12),
 c blob(12),
 KEY  (`a`)
@@ -1501,7 +1501,7 @@
 DROP TABLE t5;
 
 CREATE TABLE `t5` (
-`a` int(11) NOT NULL auto_increment,
+`a` int NOT NULL auto_increment,
 b varchar(250),
 c varchar(800),
 KEY  (`a`)
@@ -1519,7 +1519,7 @@
 SELECT * FROM t5;
 
 CREATE TABLE `t6` (
-`a` int(11) NOT NULL auto_increment,
+`a` int NOT NULL auto_increment,
 b blob(12),
 c int,
 KEY  (`a`)

=== modified file 'tests/t/disabled.def'
--- tests/t/disabled.def	2008-07-19 06:08:40 +0000
+++ tests/t/disabled.def	2008-08-05 06:19:05 +0000
@@ -26,7 +26,7 @@
 
 subselect_notembedded  : Bug#35803, Bug#33506 2008-Apr-03 subselect_notembedded crashes the server
 #alter_table : Bug introduced around datatime
-archive: Something is up in crashed table
+#archive: Something is up in crashed table
 archive_aio_posix: Something is up in crashed table
 archive_bitfield: Something is up in crashed table
 auto_increment: Duplicate key on primary error (possible test fix)
@@ -60,9 +60,6 @@
 func_str: strict mode, rewrite test
 func_test: remove old style || 
 func_time: strict mode, rewrite test
-heap: weird FK problem
-heap_btree: weird FK problem
-heap_hash: weird FK problem
 init_connect: bug in not executing sql on connect from variable
 init_file: Unknown init bug
 innodb_mysql: Foreign Key issue

=== modified file 'tests/t/heap.test'
--- tests/t/heap.test	2008-06-25 05:29:13 +0000
+++ tests/t/heap.test	2008-08-05 06:19:05 +0000
@@ -155,7 +155,7 @@
 # Bug#4411 Server hangs when trying to SELECT MAX(id) from an empty HEAP table
 #
 CREATE TABLE `job_titles` (
-  `job_title_id` int(6) unsigned NOT NULL default '0',
+  `job_title_id` int unsigned NOT NULL default '0',
   `job_title` char(18) NOT NULL default '',
   PRIMARY KEY  (`job_title_id`),
   UNIQUE KEY `job_title_id` (`job_title_id`,`job_title`)
@@ -179,7 +179,7 @@
 # Bug #6748
 # heap_rfirst() doesn't work (and never did!)
 #
-CREATE TABLE t1 (pseudo char(35) PRIMARY KEY, date int(10) unsigned NOT NULL) ENGINE=HEAP;
+CREATE TABLE t1 (pseudo char(35) PRIMARY KEY, date int unsigned NOT NULL) ENGINE=HEAP;
 INSERT INTO t1 VALUES ('massecot',1101106491),('altec',1101106492),('stitch+',1101106304),('Seb Corgan',1101106305),('beerfilou',1101106263),('flaker',1101106529),('joce8',5),('M4vrick',1101106418),('gabay008',1101106525),('Vamp irX',1101106291),('ZoomZip',1101106546),('rip666',1101106502),('CBP ',1101106397),('guezpard',1101106496);
 DELETE FROM t1 WHERE date<1101106546;
 SELECT * FROM t1;
@@ -224,7 +224,7 @@
 show create table t1;
 alter table t1 modify v char(10);
 show create table t1;
-alter table t1 modify t varchar(10);
+alter table t1 modify t varchar(50);
 show create table t1;
 select concat('*',v,'*',c,'*',t,'*') from t1;
 drop table t1,t2,t3;

=== added file 'tests/t/heap_var.test'
--- tests/t/heap_var.test	1970-01-01 00:00:00 +0000
+++ tests/t/heap_var.test	2008-08-05 05:01:04 +0000
@@ -0,0 +1,78 @@
+#
+# Test heap tables with variable-sized records.
+#
+
+--disable_warnings
+drop table if exists t1;
+--enable_warnings
+
+set @@session.max_heap_table_size=16*1024*1024;
+
+--error 1234
+create table t1 (a int not null, b varchar(400), c int, primary key (a), key (c)) engine=heap comment="testing heaps" block_size=128;
+
+--error 1234
+create table t1 (a int not null, b int, c varchar(400), primary key (a), key (b)) engine=heap comment="testing heaps" block_size=4;
+
+create table t1 (a int not null, b int, c varchar(400), d varchar(400), primary key (a), key (b)) engine=heap comment="testing heaps" block_size=24;
+
+#show table status like "t1";
+
+insert into t1 values (1,1,'012',NULL), (2,2,'0123456789',NULL), (3,3,'012345678901234567890123456789',NULL), (4,4,NULL,'0123456789012345678901234567890123456789012345678901234567890123456789');
+select * from t1;
+
+delete from t1 where a = 3;
+select * from t1;
+
+insert into t1 values (5,5,NULL,'0123'), (6,6,NULL,'0123');
+select * from t1;
+
+update t1 set c = '012345678901234567890123456789' where a = 2;
+select * from t1;
+
+update t1 set c = '0123456789' where a = 2;
+select * from t1;
+
+insert into t1 values (7,7,'0123',NULL), (8,8,'0123',NULL);
+select * from t1;
+
+#show table status like "t1";
+alter table t1 block_size = 0;
+#show table status like "t1";
+alter table t1 row_format = dynamic;
+#show table status like "t1";
+alter table t1 block_size = 128, max_rows = 10001;
+#show table status like "t1";
+
+select * from t1;
+
+delete from t1;
+select * from t1;
+
+let $1=10001;
+
+disable_query_log;
+
+while ($1) 
+{
+
+  eval insert into t1 values ($1,$1,$1,$1);
+
+  dec $1;
+
+}
+enable_query_log;
+
+select count(*) from t1;
+
+--error 1114
+insert into t1 values (100000,100000,NULL,'0123'), (100000,100000,NULL,'0123');
+
+#show table status like "t1";
+select count(*) from t1;
+
+set @@session.max_heap_table_size=default;
+
+drop table t1;
+
+# End of 5.0 tests

